from pyhipp.io import h5
from pyhipp.core import abc
from pyhipp_sims import sims
import numpy as np
from functools import cached_property


class ReconstAreaMask(abc.HasDictRepr):

    repr_attr_keys = ('sim_info',)

    def __init__(self, sim_info: str | sims.SimInfo, **base_kw) -> None:
        super().__init__(**base_kw)

        if isinstance(sim_info, str):
            sim_info = sims.predefined[sim_info]

        self.sim_info: sims.SimInfo = sim_info

    def mask_value_at(self, x_sim: np.ndarray) -> np.ndarray:
        raise NotImplementedError()

    def is_in_reconst_area(self, x_sim: np.ndarray) -> np.ndarray:
        raise NotImplementedError()


class ReconstAreaMaskSdssL500(ReconstAreaMask):

    repr_attr_keys = ReconstAreaMask.repr_attr_keys + ('f_selected',)

    def __init__(self, sim_info: str | sims.SimInfo = 'elucid',
                 mask_name='SDSS_reconstruction_area',
                 **base_kw) -> None:
        '''
        @mask_name: name of the mask in the supplementary data. Allowed 
        values: 
        - 'SDSS_reconstruction_area'.
        - 'SDSS_reconstruction_area_large'.
        '''

        super().__init__(sim_info=sim_info, **base_kw)

        sim_info = self.sim_info
        p_meta = (sim_info.root_dir / 'supplementary' /
                  'reconstruction_pipeline' / 'geometry.hdf5')
        mask = h5.File.load_from(p_meta, f'Masks/{mask_name}') > .9

        self.mask_reconst_area = mask

        L = sim_info.box_size
        N = mask.shape[0]
        assert mask.shape == (N, N, N)

        self.box_size = L
        self.n_grids = N
        self.cell_size = L/N
        self.f_selected = mask.sum() / mask.size

    def mask_value_at(self, x_sim: np.ndarray) -> np.ndarray:
        '''
        @x_sim: position(s) in the simulation frame. May be shifted (out-of-box) 
        at most one box size.
        '''
        N, L_cell = self.n_grids, self.cell_size
        mask = self.mask_reconst_area

        x_sim = np.asarray(x_sim)
        cid = np.floor(x_sim / L_cell + 0.5).astype(int)
        cid[cid < 0] += N
        cid[cid >= N] -= N

        assert np.all(cid >= 0)
        assert np.all(cid <= N)

        return mask[tuple(cid.T)]

    def is_in_reconst_area(self, x_sim: np.ndarray) -> np.ndarray:
        return self.mask_value_at(x_sim)


class _ReconstAreaMasks:
    @cached_property
    def small(self):
        return ReconstAreaMaskSdssL500(mask_name='SDSS_reconstruction_area')

    @cached_property
    def large(self):
        return ReconstAreaMaskSdssL500(
            mask_name='SDSS_reconstruction_area_large')

    @cached_property
    def small_n512(self):
        return ReconstAreaMaskSdssL500(
            mask_name='SDSS_reconstruction_area_N512')

    @cached_property
    def large_n512(self):
        return ReconstAreaMaskSdssL500(
            mask_name='SDSS_reconstruction_area_large_N512')


reconst_area_masks = _ReconstAreaMasks()



from typing import Any, Iterable, List
from pyhipp.core import abc
from dataclasses import dataclass
import numpy as np


class Job(abc.HasDictRepr):
    pass
        
class IndexedJob(Job):
    
    repr_attr_keys = ('id',)
    
    def __init__(self, id: int, **kw) -> None:
        super().__init__(**kw)
        self.id = id
    

class JobList(abc.HasDictRepr):
    
    JobType: type = Job
    
    def __init__(self, **kw) -> None:
        super().__init__(**kw)
    
    def __iter__(self) -> Iterable[JobType]:
        raise NotImplementedError()
        
class IndexedJobList(JobList):
    
    JobType: type = IndexedJob
    repr_attr_keys = ('id_list', 'size')
    
    def __init__(self, id_list: List[int],  **kw) -> None:
        super().__init__(**kw)
        self.id_list = id_list
    
    @property
    def size(self) -> int:
        return len(self.id_list)
    
    def __iter__(self) -> Iterable[JobType]:
        JobType = self.JobType
        for id in self.id_list:
            yield JobType(id)
    
    def __len__(self) -> int:
        return self.size
    
    def __getitem__(self, i: int) -> JobType:
        return self.JobType(self.id_list[i])

class Scheduler(abc.HasDictRepr):
    def __init__(self, **kw) -> None:
        super().__init__(**kw)
        
class StaticScheduler(Scheduler):
    '''
    n_jobs and the the jobs of each worker are determined at the construction 
    time.
    '''
    
    repr_attr_keys = ('n_jobs', 'n_workers', 'pl')
    
    @dataclass
    class Policy:
        batch_size: int     = 0
    
    def __init__(self, n_jobs: int, n_workers: int, pl: Policy = None,
                 **kw) -> None:
        
        super().__init__(**kw)
        
        assert n_workers > 0
        if pl is None:
            pl = StaticScheduler.Policy()
            
        self.n_jobs = n_jobs
        self.n_workers = n_workers
        self.pl = pl
        
    @property
    def used_batch_size(self):
        bsz = self.pl.batch_size
        if bsz > 0:
            return bsz
        n_jobs, n_workers = self.n_jobs, self.n_workers
        return (n_jobs + n_workers - 1) // self.n_workers
            
    def jobs(self, worker_id: int) -> IndexedJobList:
        '''
        @worker_id: an index in [0, n_workers). Negative values are allowed.
        '''
        bsz, n_jobs, n_workers = (self.used_batch_size, 
                                  self.n_jobs, self.n_workers)
        group_bsz = n_workers * bsz
        if worker_id < 0:
            worker_id = n_workers + worker_id
        assert worker_id < n_workers and worker_id >= 0
        
        job_id_list = []
        group_b = 0
        while group_b < n_jobs:
            group_e = group_b + group_bsz
            if group_e <= n_jobs:    
                b = group_b + worker_id * bsz   
                e = b + bsz
            else:
                n_jobs_left = n_jobs - group_b
                bsz = n_jobs_left // n_workers
                n_jobs_left -= bsz * n_workers
                if worker_id < n_jobs_left:
                    b = group_b + worker_id * (bsz + 1)
                    e = b + bsz + 1
                else:
                    b = group_b + n_jobs_left + worker_id * bsz
                    e = b + bsz
            job_id_list.append(np.arange(b, e))
            group_b = group_e
            
        return IndexedJobList(np.concatenate(job_id_list))
        
        
        


from __future__ import annotations
from pyhipp.core import abc
from typing import Tuple, Union
import numpy as np
from numba import njit

class NonPeriodicField(abc.HasDictRepr):
    
    Shape = Tuple[int, ...]
    Axes = Tuple[int, ...]
    DType = np.dtype
    
    def __init__(self, data: np.ndarray, 
                 spatial_n_dims: int = 3, **kw) -> None:
        
        super().__init__(**kw)
        
        self.data = data
        self.spatial_n_dims = spatial_n_dims
        
    @classmethod
    def from_shape(cls, shape: Shape, dtype: DType = np.float64, 
                   fill = 0, **init_kw):
        if fill is not None:
            if fill == 0:
                data = np.zeros(shape, dtype=dtype)
            else:
                data = np.full(shape, fill, dtype=dtype)
        else:
            data = np.empty(shape, dtype=dtype)
        return cls(data=data, **init_kw)
            
    @property
    def dtype(self) -> np.dtype:
        return self.data.dtype
    
    @property
    def n_dims(self) -> int:
        return self.data.ndim
    
    @property
    def shape(self) -> Shape:
        return self.data.shape
    
    @property
    def elem_shape(self) -> Shape:
        return self.shape[self.spatial_n_dims:]
    
    @property
    def spatial_axes(self) -> Axes:
        return tuple(dim for dim in range(self.spatial_n_dims))
    
    def to_simple_repr(self) -> dict:
        out = super().to_simple_repr()
        out |= {
            'shape': self.shape,
            'dtype': self.dtype,
            'spatial_n_dims': self.spatial_n_dims
        }
        return out
    
class NonPeriodicCIC(abc.HasDictRepr, abc.HasLog):
    
    repr_attr_keys = ('x_range', 'n_grids', 'l_grid', 'field')
    
    Field = NonPeriodicField
    
    def __init__(self, 
                 x_range: np.ndarray, 
                 n_grids: Union[np.ndarray, int] = 10,
                 dtype = np.float64,
                 field: Field = None,
                 **kw) -> None:
        '''
        @x_range: ndarray of shape (3, 2), specifying the lower and upper bounds
                  of three dimensions, or of shape (2,), the bounds of every 
                  dimension.
        @n_grids: the numbers of grid points of three dimensions. In each 
                  dimension, the first and last grid points are at the lower 
                  and upper bounds, respectively.
        '''
        super().__init__(**kw)
        
        if np.isscalar(n_grids):
            n_grids = n_grids, n_grids, n_grids
        n_grids = np.array(n_grids, dtype=int)
        assert n_grids.shape == (3,)
        n_cells = n_grids - 1
        
        x_range = np.array(x_range, dtype=float)
        if x_range.ndim == 1:
            x_range = np.array([x_range, x_range, x_range])
        assert x_range.shape == (3, 2)
        
        self.log(f'Field data shape = {n_grids}')
        if field is None:
            field = NonPeriodicCIC.Field.from_shape(
                n_grids, dtype=dtype)
        assert field.shape == tuple(n_grids)
        
        x_lb, x_ub = x_range.T
        l_grid = (x_ub - x_lb) / n_cells
        
        self.x_range = x_range
        self.n_cells = n_cells
        self.n_grids = n_grids
        self.x_lb, self.x_ub, self.l_grid = x_lb, x_ub, l_grid
        self.field = field
        
    def add(self, x: np.ndarray, weight: np.ndarray = None):
        field = self.field
        
        assert x.ndim == 2
        n_pts = len(x)
        if weight is not None:
            assert weight.shape == (n_pts, )
            
        x = x - self.x_lb
        cid_l = np.floor(x / self.l_grid).astype(np.int64)
        sel = ((cid_l >= 0) & (cid_l < self.n_cells)).all(1)
        cid_l, x = cid_l[sel], x[sel]
        if weight is not None:
            weight = weight[sel]
        n_pts_sel = len(x)
        self.log(f'Add {n_pts_sel}/{n_pts} points to grids')
        
        cid_r = cid_l + 1
        cid_lr = cid_l, cid_r
        x_rel = x / self.l_grid
        w_lr = cid_r - x_rel, x_rel - cid_l
        
        for i0 in 0,1:
            for i1 in 0,1:
                for i2 in 0,1:
                    cid = cid_lr[i0][:,0], cid_lr[i1][:,1], cid_lr[i2][:,2]
                    w = w_lr[i0][:, 0] * w_lr[i1][:, 1] * w_lr[i2][:, 2]
                    if weight is not None:
                        w *= weight
                    self.__add_to(field.data, *cid, w)
    
    @staticmethod
    @njit
    def _add_to(field: np.ndarray, cid0: np.ndarray, 
                 cid1: np.ndarray, cid2: np.ndarray, w: np.ndarray):
        _, n1, n2 = field.shape
        field = field.ravel()
        for i in range(len(w)):
            _cid0, _cid1, _cid2, _w = cid0[i], cid1[i], cid2[i], w[i]
            off = (_cid0 * n1 + _cid1) * n2 + _cid2
            field[off] += _w


from .cell import Cells, CelledHalos, CelledMaxHaloRadii, _Cells, _CelledHalos, _CelledMaxHaloRadii
from .domain_locator import _DomainLocatorDetail, DomainLocatorDetail, Policy, DomainLocator
from .domain_profile import DomainProfile, DomainProfileInterpolator
from .domain_sampler import DomainSampler


from __future__ import annotations
import numpy as np
from typing import Union, Sequence, Tuple
from .fft import FFTPolicy
    
class Field:

    Shape = Tuple[int, ...]
    
    class FFTContext:
        def __init__(self, policy: FFTPolicy, field: Field) -> None:
            self.policy = policy
            self.field = field
            
        @property
        def field_type(self):
            return type(self.field)
        
        @property
        def forward(self) -> Field:
            pl, field = self.policy, self.field
            data = pl.replaced(axes=field.axes).forward(field.data)
            return self.field_type(data = data, fft_policy=self.policy)
        
        @property
        def backward(self) -> Field:
            pl, field = self.policy, self.field
            data = pl.replaced(axes=field.axes).backward(field.data)
            return self.field_type(data = data, fft_policy=self.policy)

    def __init__(self, data: np.ndarray, fft_policy: FFTPolicy = None) -> None:
        
        self.data = data
        
        if fft_policy is None:
            fft_policy = FFTPolicy()
            
        self.fft_policy = fft_policy
        
    @property
    def fft(self):
        return type(self).FFTContext(self.fft_policy, self)
    
    @property
    def dtype(self) -> np.dtype:
        return self.data.dtype
    
    @property
    def data_n_dims(self) -> int:
        return self.data.ndim
    
    @property
    def data_shape(self) -> Field.Shape:
        return self.data.shape

    @property
    def n_dims(self) -> int:
        raise NotImplementedError()
    
    @property
    def shape(self) -> Field.Shape:
        raise NotImplementedError()
    
    @property
    def axes(self) -> Tuple[int,...]:
        return np.arange(self.n_dims)
    
    @property
    def imag(self) -> Field:
        return type(self)(data=self.data.imag, fft_policy=self.fft_policy)
    
    @property
    def real(self) -> Field:
        return type(self)(data=self.data.real, fft_policy=self.fft_policy)
    
    def __repr__(self) -> str:
        data_shape, dtype = self.data_shape, self.dtype
        name = type(self).__name__
        return f'{name}({data_shape=}, {dtype=})'


class ScalarField(Field):
    @property
    def n_dims(self) -> int:
        return self.data.ndim
            
    @property
    def shape(self) -> Field.Shape:
        return self.data.shape
    
    def __repr__(self) -> str:
        shape, dtype = self.shape, self.dtype
        name = type(self).__name__
        return f'{name}({shape=}, {dtype=})'

class ScalarField3D(ScalarField):
    def __init__(self, 
                 shape: Union[int, Field.Shape] = None, 
                 dtype = np.float64, 
                 data: np.ndarray = None, **kw) -> None:
        if data is None:
            if np.isscalar(shape):
                shape = (shape,) * 3
            assert len(shape) == 3
            data = np.zeros(shape, dtype=dtype)
        
        super().__init__(data, **kw)
        

class VectorField(Field):
    
    ScalarField = ScalarField    
    
    @classmethod
    def from_comps(cls, comps: Sequence[ScalarField]):
        return cls(data=np.stack([comp.data for comp in comps], axis=-1))
    
    @property
    def n_dims(self) -> int:
        return self.data.ndim - 1
    
    @property
    def shape(self) -> Field.Shape:
        return self.data.shape[:-1]
    
    @property
    def n_comps(self) -> int:
        return self.data.shape[-1]
    
    def comp(self, i: int):
        return type(self).ScalarField(data=self.data[...,i], 
                fft_policy=self.fft_policy)
    
    @property
    def comps(self):
        return (self.comp(i) for i in range(self.n_comps))
        
    def __repr__(self) -> str:
        shape, n_comps, dtype = self.n_comps, self.shape, self.dtype
        name = type(self).__name__
        return f'{name}({shape=}, {n_comps=}, {dtype=})'
    
class VectorField3D(VectorField):

    ScalarField = ScalarField3D

    def __init__(self, 
                 shape: Union[int, Field.Shape] = None, 
                 n_comps = 3,
                 dtype = np.float64,
                 data: np.ndarray = None, **kw) -> None:
        if data is None:
            if np.isscalar(shape):
                shape = (shape,) * 3
            assert len(shape) == 3
            data = np.zeros((*shape, n_comps), dtype=dtype)
            
        super().__init__(data, **kw)
    



from __future__ import annotations
from .field import Field
from .mesh import Mesh
import numpy as np
from numba.experimental import jitclass
from numba import njit

class MassAssignment:
    
    def __init__(self, mesh: Mesh) -> None:
    
        self.mesh = mesh
        
    @staticmethod
    def from_method(self, method='CIC') -> MassAssignment:
        return {
            'CIC': CIC
        }[method]
        
    def compensate_shape(self, field: Field):
        raise NotImplementedError()

@jitclass
class _CIC_Auxiliary:
    
    def __init__(self) -> None:
        pass
    
    @staticmethod
    def window_at_x(x: np.ndarray):
        '''
        @x: arbitrary shaped.
        '''
        w = 1.0 - np.abs(x)
        w = np.clip(w, 0.0, None)
        return w
    
    @staticmethod
    def window_at_ik(ik: np.ndarray, n_grids: int):
        '''
        @ik: integer indices of the grid points in k-space, arbitrary shaped.
        '''
        sinc = np.sinc(ik / n_grids)
        w = sinc * sinc
        return w
    
    @staticmethod
    def find_near(x: np.ndarray, n_grids: int):
        x_l = np.floor(x)
        w_l = 1.0 - (x - x_l)
        w_r = 1.0 - w_l
        
        ix_l = x_l.astype(np.int64)
        ix_r = ix_l + 1
        ix_r[ix_r == n_grids] = 0
        
        return (ix_l, ix_r), (w_l, w_r)
    
    @staticmethod
    def assign_point(field: np.ndarray, ix_lr, w_lr):
        for i0 in range(2):
            for i1 in range(2):
                for i2 in range(2):
                    field[..., ix_lr[i0][0], ix_lr[i1][1], ix_lr[i2][2] ] \
                        += w_lr[i0][0] * w_lr[i1][1] * w_lr[i2][2]
                        
    @staticmethod
    def assign_point_weighted(field: np.ndarray, ix_lr, w_lr, w):
        for i0 in range(2):
            for i1 in range(2):
                for i2 in range(2):
                    field[ix_lr[i0][0], ix_lr[i1][1], ix_lr[i2][2], ...] \
                        += w * w_lr[i0][0] * w_lr[i1][1] * w_lr[i2][2]
       
class CIC(MassAssignment):
       
    def to_field(self, xs: np.ndarray, field: Field, 
                 weights: np.ndarray = None):
        '''
        @xs: shaped (n_points, 3)
        @weights: None | array shaped (n_points, ) or (n_points, n_comps).
        @field: ScalarField or VectorField, shape = (n, n, n). 
        '''
        mesh = self.mesh
        n, h = mesh.n_grids, mesh.l_grid
        assert field.shape == (n, n, n)
        xs = xs / h
        xs[ xs < 0. ] += n
        xs[ xs >= n ] -= n
        assert np.all( (xs >= 0.) & (xs < n) )

        self.__assign_points(xs, weights, field.data, n)
        
    def compensate_shape(self, field: Field):
        mesh = self.mesh
        n = mesh.n_grids
        
        variant = field.fft_policy.variant
        if variant == 'real':
            assert field.shape == (n, n, n//2+1)
            is_real = True
        elif variant == 'complex':
            assert field.shape == (n, n, n)    
            is_real = False
        else:
            raise KeyError(f'Unknown variant: {variant}')    

        self.__div_shape_fn(field.data, n, is_real)
    
    @staticmethod
    @njit
    def __div_shape_fn(field: np.ndarray, n_grids: int, is_real: bool):
        aux = _CIC_Auxiliary()
        n_half = n_grids // 2
        n0, n1 = n_grids, n_grids
        n2 = n_half + 1 if is_real else n_grids
        for i0 in range(n0):
            ik0 = i0 - n_grids if i0 > n_half else i0
            for i1 in range(n1): 
                ik1 = i1 - n_grids if i1 > n_half else i1
                for i2 in range(n2):
                    ik2 = i2 - n_grids if i2 > n_half else i2
                    w = aux.window_at_ik(ik0, n_grids) * \
                        aux.window_at_ik(ik1, n_grids) * \
                        aux.window_at_ik(ik2, n_grids)
                    field[i0,i1,i2] *= 1.0 / w
                
    @staticmethod
    @njit
    def __assign_points(xs: np.ndarray, weights: np.ndarray, field: np.ndarray, 
                n_grids: int):
        aux = _CIC_Auxiliary()
        for i in range(xs.shape[0]):
            ix_lr, w_lr = aux.find_near(xs[i], n_grids)
            if weights is not None:
                aux.assign_point_weighted(field, ix_lr, w_lr, weights[i])
            else:
                aux.assign_point(field, ix_lr, w_lr)


import scipy.fft as scipy_fft
from dataclasses import dataclass, replace
import numpy as np
from typing import Tuple

@dataclass
class FFTPolicy:
    
    '''
    @shape: shape of the input array. If larger than input, pad input with 
            zeros; if smaller, crop.
    @axes: target axes to transform. Default: all axes, or last len(shape) axes 
           if specified shape. A repeated axis will be transformed multiple times.
    @norm: 'forward' | 'backward' | 'ortho' | None (means 'backward').
    @overwrite_input: whether input can be overwritten as temporary space.
    @n_workers: number of workers for parallel computation. Negative for a wrap 
                around from os.cpu_count().
    @variant: 'real' | 'complex'
    '''
    
    shape: Tuple[int,...]   = None
    axes: Tuple[int,...]    = None
    norm: str               = None
    overwrite_input: bool   = False
    n_workers: int          = None
    variant: str            = 'real'
    
    def forward(self, x: np.ndarray) -> np.ndarray:    
        fn = scipy_fft.rfftn if self.variant == 'real' else scipy_fft.fftn
        return fn(x, **self.__fftn_kw)
        
    def backward(self, x: np.ndarray) -> np.ndarray:
        fn = scipy_fft.irfftn if self.variant == 'real' else scipy_fft.ifftn
        return fn(x, **self.__fftn_kw)
    
    def replaced(self, **changes):
        return replace(self, **changes)
    
    @property
    def __fftn_kw(self):
        return dict(s=self.shape, axes=self.axes, norm=self.norm, 
            overwrite_x=self.overwrite_input, workers=self.n_workers)
        
        
default_fft_policy = FFTPolicy()


from __future__ import annotations
from typing import Tuple, Iterable, Any
import numpy as np
from numba import int64, float64, njit
from numba.experimental import jitclass
import pandas as pd
from pyhipp.core import abc, dataproc as dp, DataDict
from pyhipp.io import h5
from pyhipp.astro.cosmology import model as cosm_model
from pyhipp.stats.binning import Bins, BinnedData
from .domain_locator import DomainLocator
from ..catalogs import SimHaloCatalog
from ..sims.cell_storage import CellIter
from pathlib import Path

class DomainProfile(abc.HasLog, abc.HasDictRepr):
    
    repr_attr_keys = ('halo_cat', 'cell_iter', 'm_key', 'lm_bins', 'r_bins', 
                      'cids_1d')
    
    def __init__(self, 
                 halo_cat: SimHaloCatalog, 
                 cell_iter: CellIter, 
                 lm_bins: Bins,
                 r_bins: Bins,
                 m_key = 'm_mean200',
                 dom_loc_init_pl: DomainLocator.Policy = None,
                 dom_loc_run_pl: DomainLocator.LocatePolicy = None,
                 **kw) -> None:
        super().__init__(**kw)

        dom_loc = DomainLocator(halo_cat, m_key=m_key, pl=dom_loc_init_pl)
        
        m_halo = halo_cat[m_key]
        lm = dp.Num.safe_lg(m_halo)
        r_data = [ BinnedData(r_bins, dtype=float, shape=(2,)) 
                   for _ in range(lm_bins.n_bins) ]
        
        self.halo_cat = halo_cat
        self.dom_loc = dom_loc
        self.dom_loc_run_pl = dom_loc_run_pl
        
        self.cell_iter = cell_iter
        
        self.m_key = m_key
        self.lm = lm
        self.lm_bins = lm_bins
        
        self.r_bins = r_bins
        self.r_data = r_data
        
        self.cids_1d = []
        
    def run(self, cids_1d: np.ndarray):
        cell_iter = self.cell_iter
        for cid_1d in cids_1d:
            self.log(f'Running on cell {cid_1d}', end=' ... ')
            data = cell_iter.load_cell(cell_iter.ld_spec_x_vol, cid_1d)
            self.__call_on_cell(*data['ps/x', 'p_meta/volume'])
        self.cids_1d.extend(cids_1d)
            
    def dump(self, group: h5.Group):
        self.log(f'Dumping meta', end=' ... ')
        sim_info = self.halo_cat.sim_info
        d = {
            'sim_name': sim_info.name,
            'm_key': self.m_key,
            'r_edges': self.r_bins.x_edges,
            'lm_edges': self.lm_bins.x_edges,
            'cids_1d': self.cids_1d,
            'm_particle': sim_info.mass_table[1],
            'lm_halos': self.lm,
        }
        group.datasets.dump(d)
        
        self.log(f'Dumping profiles', end=' ... ', named=False)
        g_profs = group.create_group('Profiles')
        for i, r_data in enumerate(self.r_data):
            n_particles, volumes = r_data.data.T
            d = {
                'n_particles': n_particles,
                'volumes': volumes,
            }
            g_profs.create_group(str(i)).datasets.dump(d)
        
        self.log('Done', named=False)
    
    @staticmethod
    def gather_dumps(groups: Iterable[h5.Group], to_group: h5.Group = None):
        for i_batch, group in enumerate(groups):
            g_profs = group['Profiles']
            if i_batch != 0:
                for i_m in range(m_n_bins):
                    d = g_profs[str(i_m)].datasets.load()
                    for k, v in d.items():
                        profs[i_m][k] += v
                continue    
            meta = group.datasets.load()
            meta.pop('cids_1d')
            m_n_bins = len(meta['lm_edges']) - 1
            assert m_n_bins == len(g_profs.keys())
            profs = [ g_profs[str(i_m)].datasets.load() 
                        for i_m in range(m_n_bins) ]
        
        m_p = meta['m_particle']
        for i in range(m_n_bins):
            n_p, vol = profs[i]['n_particles', 'volumes']
            rho = dp.Num.safe_div(n_p * m_p, vol)
            profs[i] |= {
                'rhos': rho,
            }
        if to_group is not None:
            to_group.datasets.dump(meta)
            g_profs = to_group.create_group('Profiles')
            for i, prof in enumerate(profs):
                g_profs.create_group(str(i)).datasets.dump(prof)
        
        meta |= {'profiles': profs}
        return meta

    def __call_on_cell(self, xs: np.ndarray, vols: np.ndarray):
        self.log(f'Locating domains for {len(xs)} particles', 
                 end=' ... ', named=False)
        hids, r_effs = self.dom_loc.locate(xs, pl=self.dom_loc_run_pl)
        
        sel = hids >= 0
        n_nosel = (~sel).sum()
        if self.verbose and n_nosel > 0:
            self.log(f'{n_nosel} particles not in halos')    
        r_effs, vols, lms = r_effs[sel], vols[sel], self.lm[hids[sel]]
        
        self.log(f'Adding to bins', end=' ... ', named=False)
        m_ids = self.lm_bins.locate(lms)
        df = pd.DataFrame({'m_id': m_ids, 'r_eff': r_effs, 'vol': vols})
        for m_id, grp in df.groupby('m_id'):
            if m_id < 0: continue
            r_effs, vols = grp['r_eff'].to_numpy(), grp['vol'].to_numpy()
            weights = np.stack([np.ones_like(r_effs), vols], axis=1)
            self.r_data[m_id].add(r_effs, weights)
        self.log('Done for the cell', named=False)
        

@jitclass
class _DomainProfileInterpolator:
    
    lms:    float64[:]
    rs:     float64[:]
    rhos:   float64[:,:]
    
    def __init__(self, lms: np.ndarray, rs: np.ndarray, 
                 rhos: np.ndarray) -> None:
        '''
        @lms: array, shape = (n_lm_bins,). 
        @rs:  array, shape = (n_r_bins,).
        @rhos: densities, shaped (n_lm_bins, n_r_bins). The returned from, e.g.,
               rho_of(), rho_profile_of(), are interpolated from rhos.
        '''
        self.lms = lms
        self.rs = rs
        self.rhos = rhos
        
    def rho_of(self, lm: float, r: float, lmid: int, rid: int) -> float:
        lms, rs, rhos = self.lms, self.rs, self.rhos
        lm_w_l, lm_w_r, lm_id_l, lm_id_r = self.find_weight(lms, lm, lmid)
        r_w_l, r_w_r, r_id_l, r_id_r = self.find_weight(rs, r, rid)
        rho_ll, rho_lr = rhos[lm_id_l, r_id_l], rhos[lm_id_l, r_id_r]
        rho_rl, rho_rr = rhos[lm_id_r, r_id_l], rhos[lm_id_r, r_id_r]
        return rho_ll * lm_w_l * r_w_l + \
            rho_lr * lm_w_l * r_w_r + \
            rho_rl * lm_w_r * r_w_l + \
            rho_rr * lm_w_r * r_w_r
            
    def rho_profile_of(self, lm: float, lmid: int) -> np.ndarray:
        lm_w_l, lm_w_r, lm_id_l, lm_id_r = self.find_weight(self.lms, lm, lmid)
        hist_l, hist_r = self.rhos[lm_id_l], self.rhos[lm_id_r]
        return lm_w_l * hist_l + lm_w_r * hist_r
        
    def find_weight(self, xs: float, x: float, xid: int):
        n_xs = len(xs)
        if xid == 0:
            w_l, w_r = 0., 1.
            xid_l, xid_r = 0, 0
        elif xid == n_xs:
            w_l, w_r = 1.0, 0.0
            xid_l, xid_r = n_xs-1, n_xs-1
        else:
            xid_l, xid_r = xid-1, xid
            x_l, x_r = xs[xid_l], xs[xid_r]
            dx = x_r - x_l
            w_l, w_r = (x_r - x) / dx, (x - x_l) / dx
        return w_l, w_r, xid_l, xid_r

class DomainProfileInterpolator(abc.HasDictRepr, abc.HasLog):
    
    Impl = _DomainProfileInterpolator
    
    repr_attr_keys = ('meta', 'impl')
    
    def __init__(self, profile_data: DataDict, **kw) -> None:
        
        super().__init__(**kw)
        
        rhos, m_key, m_halo_lb, z, \
        lm_edges, r_edges, cosm = profile_data[
            'profiles', 'm_key', 'm_halo_lb', 'z', 
            'lm_edges', 'r_edges', 'cosmology']
        m_key = m_key.decode()
        rho_mean = cosm_model.predefined[cosm.decode()].rho_matter(z)
        rhos /= rho_mean
        
        lms = .5 * (lm_edges[1:] + lm_edges[:-1])
        lms[-1] = lm_edges[-2]
        rs = .5 * (r_edges[1:] + r_edges[:-1])
        impl = _DomainProfileInterpolator(lms, rs, rhos)
        
        lm_range = lm_edges.min(), lm_edges.max()
        r_range = r_edges.min(), r_edges.max()
        
        vol_edges = 4. / 3. * np.pi * r_edges**3
        vol_bins = vol_edges[1:] - vol_edges[:-1]
        
        self.impl = impl
        self.r_edges, self.lm_edges = r_edges, lm_edges
        self.vol_edges = vol_edges
        self.vol_bins = vol_bins
        self.meta = DataDict({
            'm_key': m_key, 'm_halo_lb': m_halo_lb, 'z': z,
            'n_lm_bins': len(lms), 'n_r_bins': len(rs),
            'lm_range': lm_range, 'r_range': r_range,
            'rho_mean': rho_mean,
        })
        
    @staticmethod
    def from_file(file_name: Path, group_name: str):
        with h5.File(file_name) as f:
            g = f[group_name]
            profile_data = g.datasets.load()
        return DomainProfileInterpolator(profile_data)
    
    def rho_of(self, lm_halo: np.ndarray, r_eff: np.ndarray):
        '''
        Return density rho / rho_mean.
        '''
        
        lm_halo, r_eff = np.asarray(lm_halo), np.asarray(r_eff)
        assert lm_halo.ndim == 1
        assert lm_halo.shape == r_eff.shape
        
        impl = self.impl
        lmid = np.searchsorted(impl.lms, lm_halo)
        rid = np.searchsorted(impl.rs, r_eff)
        rho = self.__rho_of(impl, lm_halo, r_eff, lmid, rid)
        
        return rho
    
    def mass_profile_of(self, lm_halo: float):
        '''
        Return mass, dm / (rho_mean * R_halo^3), in each radial bin.
        '''
        impl = self.impl
        lmid = np.searchsorted(impl.lms, lm_halo)
        rhos = impl.rho_profile_of(lm_halo, lmid)
        masses = rhos * self.vol_bins
        return masses
        
    @staticmethod
    @njit
    def __rho_of(impl: _DomainProfileInterpolator, 
                 lm: np.ndarray, r: np.ndarray, 
                 lmid: np.ndarray, rid: np.ndarray):
        out = np.empty_like(lm)
        for i in range(len(lm)):
            out[i] = impl.rho_of(lm[i], r[i], lmid[i], rid[i])
        return out

    


from scipy.spatial import Delaunay
from pyhipp.core import abc
from functools import cached_property
from numba import njit
import numpy as np
from pyhipp.io import h5
from .cell_storage import CellIter

class Triangulation(abc.HasLog):
    def __init__(self, points: np.ndarray, **kw) -> None:
        super().__init__()
        
        tri = Delaunay(points)
        assert len(tri.coplanar) == 0
        points = tri.points
        n_dims = points.shape[1]
        n_points = len(points)
        
        self.tri = tri
        self.points = points
        self.n_dims = n_dims
        self.n_points = n_points
        
    @cached_property
    def simplex_volumes(self):
        _pts = self.points[self.tri.simplices]
        dx = _pts[:, 1:] - _pts[:, [0]]
        vols = np.abs(np.linalg.det(dx)) * (1./6.)
        return vols
    
    @cached_property
    def vertex_volumes(self):
        vols = np.zeros(self.n_points, dtype=float)
        i_pts, simp_vols = self.tri.simplices, self.simplex_volumes
        self.__assign_volume(i_pts, simp_vols, vols)
        return vols
    
    @staticmethod
    @njit
    def __assign_volume(i_pts: np.ndarray, simp_vols: np.ndarray, 
                        vols: np.ndarray):
        n_simps, n_verts = i_pts.shape
        for i in range(n_simps):
            vols[i_pts[i]] += simp_vols[i] / n_verts
            
class ParticleVolume(abc.HasLog):
    def __init__(self, cell_iter: CellIter, **kw) -> None:
        super().__init__(**kw)
        self.cell_iter = cell_iter
        
    def run(self, boundary_thick=10, crange=None, fout_suffix=None):
        self.log('Start running')
            
        cell_iter = self.cell_iter
        snap = cell_iter.snap
        snap_dir = cell_iter.sim_info_m.snap_dir(snap)
        fout_name = snap_dir / 'particle_meta' / ('chunk.hdf5' + fout_suffix)
        
        with h5.File(fout_name, 'w') as f:
            
            g_cells = f.create_group('Cells')
            iter_x = cell_iter.iter_x_with_ngbs(
                boundary_thick=boundary_thick, crange=crange)
            for cid, x, mask in iter_x:
                
                n_p, n_in_cell = x.shape[0], mask.sum()
                self.log(f'Cell {cid}: {n_p=} {n_in_cell=}')
                
                vols = Triangulation(x).vertex_volumes
                assert len(vols) == len(x)
                
                g_cells.create_group(str(cid)).datasets.dump({
                    'volume': vols[mask]
                })
                
        self.log('Finished')


from __future__ import annotations
from pyhipp.core import abc
from typing import Any, Dict, Tuple, List, Mapping, Union, Set, Sequence, Callable
import numpy as np
from functools import cached_property

class SelectedArgs(abc.HasSimpleRepr, abc.HasName):
    def __init__(self, mask: np.ndarray = None, 
                 ids: np.ndarray = None, size: int = None,
        **kw) -> None:
        
        super().__init__(**kw)
        
        self._mask = mask
        self._ids = ids
        self._size = size
        
    @classmethod
    def create_all(self, size: int):
        return SelectedArgs(mask = np.ones(size, dtype=bool))
    
    @classmethod
    def create_none(self, size: int):
        return SelectedArgs(mask = np.zeros(size, dtype=bool))
        
    def to_simple_repr(self) -> dict:
        return {
            'name': self.name,
            'mask': self.mask,
            'ids':  self.ids,
            'size': self.size,
        }
    
    @property
    def mask(self):
        mask = self._mask
        if mask is None:
            mask = self.ids_to_mask(self._ids, self._size)
            self._mask = mask
        return mask
    
    @property
    def ids(self):
        ids = self._ids
        if ids is None:
            mask = self._mask
            ids = self.mask_to_ids(mask)
            self._ids = ids
            self._size = len(mask)
        return ids
    
    @property
    def size(self):
        size = self._size
        if size is None:
            size = len(self._mask)
            self._size = size
        return size
    
    @cached_property
    def selected_size(self):
        if self._mask is not None:
            return self._mask.sum()
        assert self._ids is not None
        return len(self._ids)
    
    @property
    def preferred_args(self):
        if self._mask is not None:
            return self._mask
        assert self._ids is not None
        return self._ids
    
    def __call__(self, *x: np.ndarray) -> Any:
        assert len(x) > 0
        args = self.preferred_args
        if len(x) == 1:
            return x[0][args]
        return tuple(_x[args] for _x in x)
    
    def __len__(self):
        return self.size
            
    @staticmethod
    def mask_to_ids(mask: np.ndarray):
        return mask.nonzero()[0]
    
    @staticmethod
    def ids_to_mask(ids: np.ndarray, size: int):
        mask = np.zeros(size, dtype=bool)
        mask[ids] = True
        return mask

class Filter(abc.HasSimpleRepr, abc.HasName):
    
    def __init__(self, **kw) -> None:
    
        super().__init__(**kw)
        
    @staticmethod
    def create_and(spec: Union[Sequence[Tuple[Any, Any]], Mapping[Any, Any]]):
        fs = []
        if isinstance(spec, Mapping):
            spec = spec.items()
        for k, v in spec:
            if np.isscalar(v) or isinstance(v, str):
                f = FilterEq(k, v)
            elif isinstance(v, Sequence):
                assert len(v) == 2
                f = FilterRange(k, v)
            elif isinstance(v, Callable):
                f = FilterByFn(k, v)
            else:
                raise ValueError(f'Invalid spec: key={k}, value={v}')
            fs.append(f)
        return FilterAnd(fs)
        
    def to_simple_repr(self) -> dict:
        return {
            'name': self.name
        }
        
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        raise NotImplementedError()
    
    def __or__(self, filter: Filter):
        return FilterOr([self, filter])
        
    def __and__(self, filter: Filter):
        return FilterAnd([self, filter])
  
    def __invert__(self):
        return FilterNegate(self)
   
        
class FilterAll(Filter):
    
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        return SelectedArgs.create_all(len(catalog))
        
class FilterNone(Filter):
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        return SelectedArgs.create_none(len(catalog))
        
class FilterEq(Filter):
    def __init__(self, key: str, value: Any, **kw) -> None:
        
        super().__init__(**kw)
        
        self.key = key
        self.value = value
        
    def to_simple_repr(self) -> dict:
        out = super().to_simple_repr()
        out['key'] = self.key
        out['value'] = self.value
        return out
        
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        return SelectedArgs(mask = catalog[self.key] == self.value)
        
class FilterNe(Filter):
    def __init__(self, key: str, value: Any, **kw) -> None:
        
        super().__init__(**kw)
        
        self.key = key
        self.value = value
        
    def to_simple_repr(self) -> dict:
        out = super().to_simple_repr()
        out['key'] = self.key
        out['value'] = self.value
        return out
        
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        return SelectedArgs(mask = catalog[self.key] != self.value)
    
class FilterRange(Filter):
    def __init__(self, key: str, range: Tuple[Any, Any], **kw) -> None:
        
        super().__init__(**kw)
        
        self.key = key
        self.range = range

    def to_simple_repr(self) -> dict:
        out = super().to_simple_repr()
        out['key'] = self.key
        out['range'] = self.range
        return out
        
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        lb, ub = self.range
        val = catalog[self.key]
        mask = (val >= lb) & (val < ub)
        return SelectedArgs(mask = mask)
    
class FilterIn(Filter):
    def __init__(self, key: str, values: Union[Tuple[Any, ...], Set[Any]], 
                 **kw) -> None:
        
        super().__init__(**kw)
        
        self.key = key
        self.values = values
        
    def to_simple_repr(self) -> dict:
        out = super().to_simple_repr()
        out['key'] = self.key
        out['values'] = self.values
        return out
    
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        mask = np.zeros(len(catalog), dtype=bool)
        vals_in = catalog[self.key]
        vals_set = self.values
        for i, val in enumerate(vals_in):
            mask[i] = val in vals_set
        return SelectedArgs(mask = mask)
        
class FilterByFn(Filter):
    def __init__(self, key: str, fn: Callable, **kw) -> None:
        super().__init__(**kw)
        
        self.key = key
        self.fn = fn
        
    def to_simple_repr(self) -> dict:
        out = super().to_simple_repr()
        out['key'] = self.key
        out['fn'] = self.fn
        return out
    
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        val = catalog[self.key]
        mask = self.fn(val)
        return SelectedArgs(mask = mask)
        
        
class FilterList(Filter):
    def __init__(self, filters: List[Filter], **kw) -> None:
        
        super().__init__(**kw)
        
        self.filters = [f for f in filters]
        
    def __len__(self):
        return len(self.filters)
    
    def to_simple_repr(self) -> dict:
        out = super().to_simple_repr()
        out['list'] = [f.to_simple_repr() for f in self.filters]
        return out
        
class FilterAnd(FilterList):
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        if len(self) == 0:
            return SelectedArgs.create_all(len(catalog))
        fs = self.filters
        f0, fs = fs[0], fs[1:]
        mask = f0(catalog).mask
        for f in fs:
            mask &= f(catalog).mask
        return SelectedArgs(mask = mask)
    
class FilterOr(FilterList):
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        if len(self) == 0:
            return SelectedArgs.create_none(len(catalog))
        fs = self.filters
        f0, fs = fs[0], fs[1:]
        mask = f0(catalog).mask
        for f in fs:
            mask |= f(catalog).mask
        return SelectedArgs(mask = mask)
    
class FilterNegate(Filter):
    def __init__(self, filter: Filter, **kw) -> None:
        
        super().__init__(**kw)
        self.filter = filter
        
    def __call__(self, catalog: Mapping) -> SelectedArgs:
        mask = self.filter(catalog).mask
        return SelectedArgs(mask = ~mask)


from __future__ import annotations
from ..gen_util import fields
from .cell_storage import CellIter
from typing import Tuple
import numpy as np
from pyhipp.core import DataDict, abc
from pyhipp.io import h5
from .sim_info_for_elucid import SimInfo, SimInfoForElucid

class PaintDensityField(abc.HasLog):
    
    def __init__(self, field: fields.PeriodicField, 
                 painter: fields.MassAssigner,
                 **kw) -> None:
        
        super().__init__(*kw)
        
        self.field = field
        self.painter = painter
        
    def paint_cells(self, cell_iter: CellIter, crange: Tuple[int,int]):
        for cell in cell_iter(cell_iter.ld_spec_x, crange=crange):
            self.paint_cell(cell)
        
    def paint_cell(self, cell: DataDict):
        x = cell['ps/x']
        self.painter.add(self.field, x)
        
    def dump(self, g: h5.Group, field_dtype=np.float32):
        field, painter = self.field, self.painter
        ccfg, data = field.cell_cfg, field.data.astype(field_dtype)
        meta = {
            'n_grids': ccfg.n_grids,
            'l_box': ccfg.l_box,
            'painter_kind': painter.kind,
            'data': data
        }
        g.datasets.dump(meta)
        
        
class SmoothDensityField(abc.HasLog):
    def __init__(self, field: fields.PeriodicField, smoother: fields.Smoother, 
                 **kw) -> None:
        
        super().__init__(**kw)
        
        self.field = field
        self.smoother = smoother
        
    def run(self) -> None:
        self.field_result = self.smoother.run(self.field)
        
    def dump(self, g: h5.Group, field_dtype=np.float32):
        data = self.field_result.data.astype(field_dtype)
        g.datasets.create('data', data, flag='ac')
        
        
class LoadDensityField(abc.HasLog, abc.HasDictRepr):
    
    repr_attr_keys = 'meta', 'field'
    
    def __init__(self, sim_info: SimInfo, snap: int, 
                 sample='Raw', n_grids=512, smooth=None, 
                 mode = 'r', dtype=np.float64,
                 **kw):
        
        super().__init__(**kw)
        
        info_e = SimInfoForElucid(sim_info)
        fname = info_e.snap_dir(snap) / 'fields.hdf5'
        
        file = h5.File(fname, mode)
        group = file[f'{sample}/NumGrids_{n_grids}/Density']
        n_grids, l_box = group.datasets['n_grids', 'l_box']
        if smooth is not None:
            group = group[smooth]
        data = group.datasets['data'].astype(dtype)
        
        n_per_grid = data.mean()
        data /= n_per_grid
        ccfg = fields.CellConfig(n_grids, l_box)
        field = fields.PeriodicField(data=data, cell_cfg=ccfg)
        
        self.sim_info = sim_info
        self.sim_info_e = info_e
        self.file = file
        self.group = group
        self.field = field                      # density contrast field
        self.meta = DataDict({
            'fname': fname, 'snap': snap, 'sample': sample, 'n_grids': n_grids,
            'smooth': smooth, 'n_per_grid': n_per_grid
        })
        
    def close(self):
        del self.group
        self.file.close()
        
    def __enter__(self):
        return self
    
    def __exit__(self, *exc):
        self.close()
        
        


from __future__ import annotations
import typing
from typing import Self
import numpy as np
import numba
from numba.experimental import jitclass
from pyhipp.field.neighbor import kd_mesh
from pyhipp.io import h5
from pyhipp_sims.sims import trees, snapshots, predefined
from pathlib import Path


def halomass2radius(m, z):
    zp1 = z + 1.
    o_m = 0.258
    o_l = 0.742
    rho_c_0 = 277536627245.83075
    e_val = np.sqrt(zp1*zp1*zp1*o_m + o_l)
    rho = rho_c_0 * e_val * e_val * 200
    r = pow(m * 3. / 4. / 3.1415 / rho, 1. / 3.) * zp1
    return r


def take_subhalos_from_tree(
        file_id: int, dst_snaps: np.ndarray, m_h_lb: float,
        o_path: Path):

    ld = trees.TreeLoaderElucidRaw(file_id)

    M200c = ld['m_crit200']
    Vmax = ld['v_max']
    Pos = ld['x']
    TreeID = ld['tree_id']
    FOFFirst = ld['f_in_grp']
    SnapNum = ld['snap']
    is_c = ld['is_c']
    z = ld.sim_info.redshifts[SnapNum]
    R200c = halomass2radius(M200c*1.0e10, z)
    FileID = np.full_like(TreeID, file_id)
    l_box = ld.sim_info.full_box_size
    Pos[Pos < 0.] += l_box
    Pos[Pos >= l_box] -= l_box
    assert np.all((Pos >= 0.) & (Pos < l_box))

    sel_0 = is_c & (M200c > m_h_lb)
    outs = {}
    for dst_snap in dst_snaps:
        sel = sel_0 & (SnapNum == dst_snap)
        outs[str(dst_snap)] = {
            'M200c': M200c[sel],
            'R200c': R200c[sel],
            'Vmax': Vmax[sel],
            'Pos': Pos[sel],
            'FileID': FileID[sel],
            'TreeID': TreeID[sel],
            'FOFFirst': FOFFirst[sel],
        }
    h5.File.dump_to(o_path, outs, f_flag='w')


def gather_subhalos_by_snap(paths, snap, o_path):
    subhalos = {}
    for p in paths:
        d = h5.File.load_from(p, key=str(snap))
        for k, v in d.items():
            subhalos.setdefault(k, []).append(v)
    subhalos = {
        k: np.concatenate(v) for k, v in subhalos.items()
    }
    h5.File.dump_to(o_path, subhalos, f_flag='w')

@jitclass
class HaloProfiles:
    xs: numba.float64[:, :]
    rs: numba.float64[:]
    n_hs: int
    l_box: float

    n_rs: int
    lr_min: float
    lr_max: float
    dlr: float

    profs: numba.float64[:, :]
    mv1: numba.float64[:]
    mcnt: numba.float64[:]

    def __init__(self, xs, rs, l_box, n_rs=20, lr_min=-2., lr_max=0.):
        assert len(xs) == len(rs)
        n_hs = len(xs)

        self.xs = xs
        self.rs = rs
        self.n_hs = n_hs
        self.l_box = l_box

        dlr = (lr_max - lr_min) / n_rs
        self.n_rs = n_rs
        self.lr_min = lr_min
        self.lr_max = lr_max
        self.dlr = dlr

        self.profs = np.zeros((n_hs, n_rs), dtype=np.float64)
        self.mv1 = np.zeros(n_hs, dtype=np.float64)
        self.mcnt = np.zeros(n_hs, dtype=np.float64)

    def on(self, i_h, xs_ngbs):
        x_h, r_h = self.xs[i_h], self.rs[i_h]
        l_box = self.l_box
        l_half = l_box * 0.5
        lr_min, dlr, n_rs = self.lr_min, self.dlr, self.n_rs
        r_max = 10.0**self.lr_max
        for x_ngb in xs_ngbs:
            dx_sq = 0.
            for dx in x_ngb - x_h:
                if dx > l_half:
                    dx -= l_box
                elif dx < -l_half:
                    dx += l_box
                dx_sq += dx * dx
            dr = np.sqrt(dx_sq)
            if dr <= r_h:
                self.mv1[i_h] += dr
                self.mcnt[i_h] += 1.0
            if dr < r_max:
                lr = np.log10(dr / r_h + 1.0e-9)
                i_r = np.int64(np.floor((lr - lr_min)/dlr))
                if i_r >= 0 and i_r < n_rs:
                    self.profs[i_h, i_r] += 1.0


def find_profile(path: Path, snap, file_ids: np.ndarray):
    info = predefined['elucid']
    l_box = info.full_box_size
    n_grids = 512
    mesh = kd_mesh._PE3Mesh(l_box, n_grids)

    with h5.File(path) as f:
        x_hs, r_hs, m_hs = f.datasets['Pos', 'R200c', 'M200c']
        x_hs, r_hs, m_hs = np.asarray(x_hs, dtype=np.float64), \
            np.asarray(r_hs, dtype=np.float64), \
            np.asarray(m_hs, dtype=np.float64)
    profs = HaloProfiles(x_hs, r_hs, l_box)

    for file_id in file_ids:
        print(f'Processing {file_id=}', flush=True)
        ld = snapshots.SnapshotLoaderElucidRaw(
            info, snap=snap, chunk_id=file_id)
        xs = np.asarray(ld.load_particles('x'), dtype=np.float64)
        xs[xs < 0.] += l_box
        xs[xs >= l_box] -= l_box
        kd = kd_mesh._PE3_from_meshing_points(mesh, xs)
        kd.query_points(profs)
        m_p = ld.part_mass_dm
    print(f'Part mass: {m_p}', flush=True)

    mcnt = profs.mcnt * m_p
    pfs = profs.profs * m_p
    mv1 = profs.mv1 * m_p / m_hs

    h5.File.dump_to(path, {
        'MassVariance1': mv1,
        'MCnt': mcnt,
        'Prof': pfs,
        # ConcFit = ld[]
    }, dump_flag='ac')



from .non_periodic import NonPeriodicField, NonPeriodicCIC
from .smoothers import Smoother, TopHatSmoother, MassAdaptiveSmoother, \
    FFTTopHatSmoother, FFTMassAdaptiveSmoother, FFTPolicy
from .periodic import PeriodicField, PeriodicCIC, CellConfig, MassAssigner 


from pyhipp.core import abc, DataDict
from pyhipp.stats.random import Rng
import numpy as np
from scipy.stats import rv_histogram
from ..catalogs import SimHaloCatalog
from .domain_profile import DomainProfileInterpolator
from .domain_locator import DomainLocator

class DomainSampler(abc.HasLog, abc.HasDictRepr):
    
    repr_attr_keys = ('n_halos', 'profile', 'm_particle', 
                      'nbins_r1', 'nbins_max')
    def __init__(self, 
        halo_cat:        SimHaloCatalog, 
        profile:         DomainProfileInterpolator, 
        m_particle:      float                      = None,
        r_eff_max:       float                      = 30.0,
        dom_loc_init_pl: DomainLocator.Policy       = None, 
        dom_loc_run_pl:  DomainLocator.LocatePolicy = None,
        seed                                        = 0, 
        **kw) -> None:
        
        super().__init__(**kw)
        
        m_key = profile.meta['m_key']
        m_halo: np.ndarray = halo_cat[m_key]
        x_halo: np.ndarray = halo_cat['x']
        assert m_key == 'm_mean200'
        r_halo: np.ndarray = halo_cat.r_halo('mean200')
        dom_loc = DomainLocator(halo_cat, m_key=m_key, pl=dom_loc_init_pl)
        
        n_bins_r1, n_bins_max = np.searchsorted(profile.r_edges, 
            [1.0001, r_eff_max+1.0e-4]) - 1
        r_edges = profile.r_edges[:n_bins_max+1]
        rng = Rng(seed)
        snap = halo_cat.meta['snap']
        if m_particle is None:
            info = halo_cat.sim_info
            m_particle = info.mass_table[info.part_type_index.dm]
        
        self.snap = snap
        self.m_halo, self.x_halo, self.r_halo = m_halo, x_halo, r_halo
        self.n_halos = len(m_halo)
        self.dom_loc = dom_loc
        self.dom_loc_run_pl = dom_loc_run_pl
        
        self.profile = profile
        self.m_particle = m_particle
        self.nbins_r1, self.nbins_max = n_bins_r1, n_bins_max
        self.r_edges = r_edges                          # bin edges of r_eff
        self.rng = rng
        
        self.log(f'Initialized. snap = {self.snap}, {m_particle=}, '
            f'r_eff bin edges = '
            f'[{r_edges[0]} ... {r_edges[n_bins_r1]} ... {r_edges[-1]}]')
        
    def set_rng(self, rng: Rng) -> None:
        self.rng = rng
        
    def sample(self, hid: int, sampling_fraction: float = 1.0):
        assert hid >= 0 and hid < len(self.m_halo)
        x_halo, m_halo, r_halo = self.x_halo[hid], \
            self.m_halo[hid], self.r_halo[hid]
        return self.sample_at(x_halo, m_halo, r_halo, 
            accept_in_domain=hid, sampling_fraction=sampling_fraction)
        
    def sample_at(self, x_halo: np.ndarray, m_halo: float, r_halo: float,
            accept_in_domain: int = None, sampling_fraction: float = 1.0):
        '''
        Give the sampled particles near one halo at `x` with mass `m_halo`.
        '''
        n_r1, n_max = self.nbins_r1, self.nbins_max
        r_edges, profile, rng = self.r_edges, self.profile, self.rng
        m_p = self.m_particle
        dom_loc, pl = self.dom_loc, self.dom_loc_run_pl
        
        masses = profile.mass_profile_of(np.log10(m_halo))[:n_max]
        m_total = masses.sum() / masses[:n_r1].sum() * m_halo
        m_sampled = m_total * sampling_fraction
        n_ps =  int(np.floor(m_sampled / m_p))
        rv, prob = rng.random(), (m_sampled - n_ps * m_p) / m_p
        if rv < prob:
            n_ps += 1
        if self.verbose:
            self.log(f'Sampling for {m_halo=}. {m_total=}, '
                     f'{sampling_fraction=}, {m_sampled=}, {n_ps=}.')

        r_effs = rv_histogram((masses, r_edges), 
            density=False, seed=rng._np_rng).rvs(size=n_ps)
        x_ps = rng.uniform_sphere(size=n_ps) * (r_effs*r_halo)[:, None] + x_halo
        if accept_in_domain is None:
            return x_ps, None
        
        n_sampled = len(x_ps)
        hids, r_effs = dom_loc.locate(x_ps, pl=pl)
        sel = hids == accept_in_domain
        x_ps, r_effs = x_ps[sel], r_effs[sel]
        n_accepted = len(x_ps)
        if self.verbose:
            self.log(f'Accepted {n_accepted} / {n_sampled} particles')
        return x_ps, r_effs

        



from pyhipp_sims.sims import SimInfo

class SimInfoForElucid:
    def __init__(self, sim_info: SimInfo) -> None:
        self.sim_info = sim_info
        
    @property
    def root_dir(self):
        return self.sim_info.root_dir / 'elucid_workdir'
    
    @property
    def root_file(self):
        return self.root_dir / 'simulation.hdf5'
    
    
    def snap_dir(self, snap: int):
        return self.root_dir / 'snapshots' / str(snap)
    
    def snap_file(self, snap: int):
        return self.snap_dir(snap) / 'snapshot.hdf5'
    
    def particle_file(self, snap: int):
        return self.snap_dir(snap) / 'particles.hdf5'


from pyhipp.core import abc
from typing import Any, Dict, Tuple
import numpy as np

class NumbaClassWrapper(abc.HasSimpleRepr):
    
    Impl: type                      = None
    impl_repr_items: Tuple          = None
    self_repr_items: Tuple          = None
    quantile_repr_items: Tuple      = None
    
    def __init__(self, impl: Impl, **kw) -> None:
        super().__init__(**kw)
        self.impl = impl
    
    @classmethod
    def create(cls, wrapper_kw: dict = None, **impl_kw):
        wrapper_kw, impl_kw = cls._parse_creator_kw(wrapper_kw, **impl_kw)
        return cls(cls.Impl(**impl_kw), **wrapper_kw)
    
    @classmethod
    def _parse_creator_kw(cls, wrapper_kw: dict = None, 
                         **kw) -> Tuple[Dict, Dict]:
        if wrapper_kw is None:
            wrapper_kw = {}
        impl_kw = {}
        for k, v in kw.items():
            if isinstance(v, NumbaClassWrapper):
                v = v.impl
            impl_kw[k] = v
        return wrapper_kw, impl_kw
    
    def to_simple_repr(self) -> dict:
        out = {
            'type': self.__class__.__name__,
        }
        items, impl = self.impl_repr_items, self.impl
        
        if items is not None: 
            for item in items:
                if isinstance(item, str):
                    out[item] = getattr(impl, item)
                else:
                    k, Wrapper = item
                    out[k] = Wrapper(getattr(impl, k)).to_simple_repr()
                
        items = self.self_repr_items
        if items is not None:
            for item in items:
                v = getattr(self, item)
                if isinstance(v, abc.HasSimpleRepr):
                    out[item] = v.to_simple_repr()
                else:
                    out[item] = repr(v)
            
        items = self.quantile_repr_items
        if items is not None:
            for item in items:
                v = out.pop(item)
                k = f'{item}(min, 0.16, 0.5, 0.84, max)'
                out[k] = self._get_repr_quantiles(v)
            
        return out
    
    @staticmethod
    def _get_repr_quantiles(values: np.ndarray) -> Tuple[np.ndarray]:
        min, max = values.min(0), values.max(0)
        qs = np.quantile(values, axis=0, q=(.16, .5, .84))
        return (min,) + tuple(qs) + (max,)


from __future__ import annotations
from typing import Any, Dict, Tuple
import numpy as np
from pyhipp.core import abc, DataDict
from pyhipp.io import h5
import pandas as pd
from pyhipp_sims.sims import SnapshotLoaderDmo, SimInfo
from .sim_info_for_elucid import SimInfoForElucid
from functools import lru_cache, cached_property
from hilbertcurve.hilbertcurve import HilbertCurve
from ..gen_util.fields import CellConfig
    
class SnapshotStorage(abc.HasLog):
    def __init__(self, ld: SnapshotLoaderDmo, n_grids: int,
                 snap_dir: str,
                 **kw) -> None:
        
        super().__init__(**kw)
        
        sim_info_e = SimInfoForElucid(ld.sim_info)
        ccfg = CellConfig(n_grids, ld.sim_info.box_size)
        
        self.ld = ld
        self.sim_info_e = sim_info_e
        self.cell_config = ccfg
        self.snap_dir = snap_dir
        
    def snapshot_file(self, mode='r'):
        snap_dir = self.sim_info_e.snap_dir(self.ld.snap)
        fname = f'{snap_dir}/particles.hdf5'
        return h5.File(fname, mode)
        
    def create_snapshot_file(self):
        ccfg, ld = self.cell_config, self.ld
        tot_n_grids = ccfg.tot_n_grids
        cell_sizes = pd.Series(np.zeros(tot_n_grids, dtype=int))
        for (x,) in ld.iter_particles('dm', ('x', )):
            cid = ccfg.cid_3d_to_1d(ccfg.x_to_cid(x))
            sizes = pd.DataFrame({'cid': cid}).groupby('cid').size()
            cell_sizes.loc[sizes.index] += sizes.to_numpy()
        cell_sizes = cell_sizes.to_numpy()
        n_parts = cell_sizes.sum()
        
        snap = ld.snap
        m_particle = ld.sim_info.mass_table[1]
        with self.snapshot_file('w') as f:
            header = f.create_group('Header')
            header.attrs.dump({
                'snap': snap,
                'total_n_particles': n_parts,
                'n_cells': ccfg.tot_n_grids,
                'm_particle': m_particle,
            })
            header.datasets.dump({
                'cell_n_particles': cell_sizes,
            })
            grp_cells = f.create_group('Cells')
            for i, n in enumerate(cell_sizes):
                grp_cell = grp_cells.create_group(str(i))._raw
                grp_cell.create_dataset('x', shape=(n, 3), dtype=np.float32)
                grp_cell.create_dataset('v', shape=(n, 3), dtype=np.float32)
                grp_cell.create_dataset('id', shape=(n,), dtype=np.int64)
        
    def push_particles(self, buffer_size = 1 * 1024 * 1024 * 1024):
        buffer_n_p = int(buffer_size / 4 / (3+3+2))
        self.log(f'Will read chunks, each with {buffer_n_p} particles.')
        ccfg, ld = self.cell_config, self.ld  
        
        with self.snapshot_file('a') as f:
            hd = f['Header']
            n_cells = hd.attrs['n_cells']
            cell_sizes = hd.datasets['cell_n_particles']
            cell_offs = np.zeros(n_cells, dtype=int)
            grp_cells = f['Cells']
            grp_cells = [ grp_cells[str(i)] for i in range(n_cells) ]
            ld_iter = ld.iter_particles(
                'dm', ('x', 'v', 'id'), n_chunk=buffer_n_p)
            for x, v, id in ld_iter:
                cid = ccfg.cid_3d_to_1d(ccfg.x_to_cid(x))
                self.__write_chunk(grp_cells, cell_offs, 
                                   cid, x, v, id)
        assert (cell_offs == cell_sizes).all()
                
    @staticmethod            
    def __write_chunk(grp_cells: list[h5.Group], cell_offs: np.ndarray, 
        cid: np.ndarray, x: np.ndarray, v: np.ndarray, id: np.ndarray):
        
        dfs = pd.DataFrame({'cid': cid}).groupby('cid')
        for _cid, df in dfs:
            cell: h5.Group = grp_cells[_cid]
            b, n = cell_offs[_cid], len(df)
            e = b + n
            cell_offs[_cid] = e
            pids = df.index.to_numpy()
            cell['x']._raw[b:e] = x[pids]
            cell['v']._raw[b:e] = v[pids]
            cell['id']._raw[b:e] = id[pids]
        n_written = cell_offs.sum()
        print(f'particles {n_written=}')
        
class CellIter(abc.HasDictRepr, abc.HasLog):
    
    '''
    sim_info_m, snap, file, cell_cfg, cache_maxcell
    header/
        snap, total_n_particles, n_cells,       -- loaded from snapshot header
        sim_name, n_side, p, l_box              -- __add_header()
    '''
    
    # used in the impl of ngb iter
    ngb_off_1d = -1,0,1
    ngb_off_3d = np.array(np.meshgrid(ngb_off_1d, ngb_off_1d, ngb_off_1d, 
                                   indexing='ij')).reshape(3,-1).T
    
    #predefiend ld_spec for __call__().
    LoadSpec = Dict[str, Tuple[str]]
    ld_spec_x = {
        'ps': ('x', )
    }
    ld_spec_x_vol = {
        **ld_spec_x,
        'p_meta': ('volume', )
    }
    
    repr_attr_keys = ('header', 'domain_sample', 'cache_maxcell')

    def __init__(self, sim_info: SimInfo, 
                 snap: int, domain_sample = None,
                 cache_maxcell = None, **kw) -> None:
        
        super().__init__(**kw)
        
        sim_info_m = SimInfoForElucid(sim_info)
        file = h5.File(sim_info_m.root_file, 'r')
                
        self.sim_info_m = sim_info_m
        self.snap = snap
        self.cache_maxcell = cache_maxcell
        
        self.file = file
        self.domain_sample = domain_sample
        self.header = self.g_ps['Header'].attrs.load()
        self.__add_header()
        
    @cached_property
    def g_snap(self):
        return self.file[f'Snapshots/{self.snap}']
        
    @cached_property
    def g_ps(self):
        samp = self.domain_sample
        if samp is None:
            group = 'Particles'
        else:
            group = f'Domain/ParticleSamples/{samp}'
        return self.g_snap[group]
    
    @cached_property
    def g_p_meta(self):
        samp = self.domain_sample
        if samp is not None:
            raise KeyError(f'Do not have meta for sample {samp}')
        return self.g_snap['ParticleMeta']
        
    def __call__(self, ld_spec: LoadSpec, 
                 crange=None) -> Any:
        for cid_1d, cid_3d in self.iter_cell_id(crange):
            out = self.load_cell(ld_spec, cid_1d)
            out |= {
                'cid_1d': cid_1d,
                'cid_3d': cid_3d,
            }
            yield out
        
    def iter_x_with_ngbs(self, boundary_thick=10.0, crange=None):
        ld = lru_cache(maxsize=self.cache_maxcell,)(CellIter.load_ps)
        ccfg = self.cell_cfg
        l_box, l_grid = ccfg.l_box, ccfg.l_grid
        l_half, l_g_half = l_box * .5, l_grid * .5
        
        for cid, cid_3d, ngb_cids, _ in self.iter_cell_id_with_ngbs(crange):
            x_all = []
            mask_all = []
            x_anc = cid_3d * l_grid + l_g_half
            x_lb = x_anc - l_g_half - boundary_thick
            x_ub = x_anc + l_g_half + boundary_thick
            
            for ngb_cid in ngb_cids:
                x = ld(self, ngb_cid, ('x', ))['x'].copy()
                
                dtype = x.dtype
                dx = x - x_anc.astype(dtype)
                x[dx >= l_half] -= l_box
                x[dx < -l_half] += l_box
                
                sel = ((x >= x_lb) & (x < x_ub)).all(1)
                if cid == ngb_cid:
                    assert sel.all()
                x = x[sel]
                
                mask = np.zeros(len(x), dtype=bool)
                mask[...] = cid == ngb_cid
                
                x_all.append(x)
                mask_all.append(mask)
        
            yield cid, np.concatenate(x_all), np.concatenate(mask_all)
  
    
    def load_rect(self, x_range: np.ndarray, ld_spec: LoadSpec):
        '''
        @x_range: ndarray like [[x_min, x_max], [y_min, y_max], [z_min, z_max]].
                  Can be slightly overbound.
        Each returned particle property is concatenated by those from visited 
        cells. All the particles in `x_range` must be returned, possibly with 
        additional neighbor particles outside `x_range`.
        '''
        x_range = np.asarray(x_range)
        ccfg = self.cell_cfg
        cid_lb = ccfg.x_to_cid_no_shift(x_range[:, 0])
        cid_ub = ccfg.x_to_cid_no_shift(x_range[:, 1])
        _out = []
        cid_list = []
        for i0 in range(cid_lb[0], cid_ub[0]+1):
            for i1 in range(cid_lb[1], cid_ub[1]+1):
                for i2 in range(cid_lb[2], cid_ub[2]+1):
                    cid = np.array((i0, i1, i2))
                    ccfg.shift_cid_in(cid)
                    ccfg.test_bound_cid(cid)
                    cid_1d = ccfg.cid_3d_to_1d(cid)
                    _out.append(self.load_cell(ld_spec, cid_1d))
                    cid_list.append(cid)
        out = DataDict({
            k: DataDict({
                k1: None for k1 in v.keys()
            }) for k, v in _out[0].items()
        })
        for k, v in out.items():
            for k1 in tuple(v.keys()):
                v[k1] = np.concatenate([ d[k][k1] for d in _out ])
        out['cid_list'] = cid_list
        return out
  
    def load_cell(self, ld_spec: LoadSpec, cid_1d: int):
        loader_map = {
            'ps': self.load_ps,
            'p_meta': self.load_p_meta,
        }
        out = DataDict({
            ld_key: loader_map[ld_key](cell_id=cid_1d, keys=keys) 
                for ld_key, keys in ld_spec.items()
        })
        return out
  
  
    def load_ps(self, cell_id: int, keys: tuple[str]):
        return self.g_ps[f'Cells/{cell_id}'].datasets.load(keys=keys)
    
    def load_p_meta(self, cell_id: int, keys: tuple[str]):
        return self.g_p_meta[f'Cells/{cell_id}'].datasets.load(keys=keys)
  
    
    def iter_cell_id(self, crange=None):
        '''
        Produce an iterator of (cid_1d, cid_3d) with Peano-Hilbert order.
        '''
        n_dims = 3
        n_side, p = self.header['n_side', 'p']
        ccfg = self.cell_cfg
        idx_ub = np.array((n_side,)*n_dims)
        
        sf_curve = HilbertCurve(p, n_dims)
        
        for i in range(2**(p*n_dims)):
            cid_3d = np.array(sf_curve.point_from_distance(i))
            if (cid_3d >= idx_ub).any():
                continue
            cid_1d = ccfg.cid_3d_to_1d(cid_3d)
            if crange is not None:
                b, e = crange
                if cid_1d < b or cid_1d >= e:
                    continue
                
            yield cid_1d, cid_3d
            
    def iter_cell_id_with_ngbs(self, crange = None):
        ccfg = self.cell_cfg
        n_side = self.header['n_side']
        for cid_1d, cid_3d in self.iter_cell_id(crange):
            ngb_cid_3d = (cid_3d + self.ngb_off_3d + n_side) % n_side
            ngb_cid_1d = ccfg.cid_3d_to_1d(ngb_cid_3d)
            yield cid_1d, cid_3d, ngb_cid_1d, ngb_cid_3d
  
        
    def __add_header(self):
        n_cells = self.header['n_cells']
        n_side = 0
        while n_side**3 < n_cells:
            n_side += 1
        assert n_side**3 == n_cells
        
        p = 0
        while 2**p < n_side:
            p += 1
            
        self.log(f'Find space filling settings: {n_side=}, {p=}')
            
        info = self.sim_info_m.sim_info
        l_box, sim_name = info.box_size, info.name
        self.cell_cfg = CellConfig(n_side, l_box)
        self.header |= {
            'n_side': n_side,
            'p': p,
            'l_box': l_box,
            'sim_name': sim_name,
        }
        
    def to_simple_repr(self) -> dict:
        return {
            'type': type(self).__name__,
            'header': self.header,
            'cell_config': self.cell_cfg.to_simple_repr(),
        }
        
    def __enter__(self):
        return self
    
    def __exit__(self, *_):
        self.file.close()
        


from __future__ import annotations
import os
from pyhipp_sims import sims
from pathlib import Path
from pyhipp import plot
from pyhipp.core import DataDict

plot.runtime_config.use_stylesheets('mathtext-it')

class ColorSets:
    dark2 = plot.ColorSeq.predefined('dark2').get_rgba()
    tab10 = plot.ColorSeq.predefined('dark2').get_rgba()
    set1 = plot.ColorSeq.predefined('set1').get_rgba()
    set2 = plot.ColorSeq.predefined('set2').get_rgba()

class ProjPaths:
    proj_dir = Path(os.environ['MAHGIC_WORK_DIR']
                    ).resolve() / 'data' / 'elucid'
    
    sims_dir = proj_dir / 'sims'
    obs_dir = proj_dir / 'obs'
    figs_dir = proj_dir / 'figures'

    @staticmethod
    def sim_dir_of(sim_info: sims.SimInfo) -> Path:
        return ProjPaths.sims_dir / sim_info.name

    @staticmethod
    def save_fig(file_name: str, **savefig_kw):
        plot.savefig(ProjPaths.figs_dir / file_name, **savefig_kw)



from numba import njit
from dataclasses import dataclass
from .field import OverdensityField, PeculiarVelocityField
from .. import gen_util as gu
from 
from . import theory
from .poisson import _FFTSolverAuxiliary
from pyhipp.core import abc
import numpy as np

class Predictor(abc.HasLog):
    def __init__(self, **kw) -> None:
        super().__init__(**kw)

class LPTPredictor(Predictor):
    '''
    Predictor using Linear Perturbation Theory (LPT).
    '''
    
    @dataclass
    class Policy:
        theory: theory.LinearPerturbationTheory
    
    def __init__(self, policy: Policy, **kw) -> None:
        
        super().__init__(**kw)
        
        self.policy = policy
        
    def predict(self, delta: OverdensityField,
                z: float):
        
        mesh = delta.mesh
        aux = _FFTSolverAuxiliary(mesh)
        n = mesh.n_grids
        n_half = n // 2
        n0, n1 = n, n
        n2 = n_half + 1
        k_base
        
        for i0 in range(n0):
            ik0 = i0 - n if i0 > n_half else i0
            for i1 in range(n1): 
                ik1 = i1 - n if i1 > n_half else i1
                for i2 in range(n2):
                    ik2 = i2 - n if i2 > n_half else i2
                    w = aux.window_at_ik(ik0, n_grids) * \
                        aux.window_at_ik(ik1, n_grids) * \
                        aux.window_at_ik(ik2, n_grids)
                    field[i0,i1,i2] *= 1.0 / w
        
        delta_k = delta.field_k.data
        v_k = v.field_k.data
        assert delta_k is not None and v_k is not None
        
        theory = self.policy.theory
        growth_f = theory.growth_rate_f(z)
        H = theory.cosm.big_hubble(z)
        a = 1.0 / (1.0 + z)
        
    def __predict(delta_k: np.ndarray, n: int, mesh: Mesh):
        
        
        
    


from __future__ import annotations
from numba.experimental import jitclass
from numba import float64, int64, bool_, njit, prange
import numba
import numpy as np
from ..gen_util import abc as abc_e
from pyhipp.core import abc
from .cell import _CelledHalos, _CelledMaxHaloRadii, _Cells, CelledHalos, Cells, CelledMaxHaloRadii
from dataclasses import dataclass
from ..catalogs import SimHaloCatalog


@jitclass
class _DomainLocatorDetail:
    src_celled_halos:                   _CelledHalos
    celled_halos:                       _CelledHalos
    big_halos:                          _CelledHalos
    hid_map:                            int64[:]        # hid -> offset in src_celled_halos (x, r, etc.)
    
    brute_force:                        bool_
    
    pre_chk_r_eff_max:                  float64
    pre_chk_n_big_halos:                int64
    pre_chk_expand:                     bool_
    pre_chk_expand_policy:              int64
    pre_chk_expand_r_eff_min:           float64
    pre_chk_expand_r_min:               float64
    
    r_max:                              float64
    celled_r_h_max:                     _CelledMaxHaloRadii
    use_celled_r_h_max:                 bool_
    celled_r_h_max_n_grids:             int64
    sort_ngb_cells:                     bool_
    exclude_cell_by_dist:               bool_

    def __init__(self, 
        celled_halos:                   _CelledHalos,
        brute_force:                    bool,
        pre_chk_r_eff_max:              float, 
        pre_chk_n_big_halos:            int,
        pre_chk_expand:                 bool,
        pre_chk_expand_policy:          int,            # 0: adjacent, 1: adjacent + diagonal
        pre_chk_expand_r_eff_min:       float,
        pre_chk_expand_r_min:           float,
        r_max:                          float,
        use_celled_r_h_max:             bool,
        celled_r_h_max_n_grids:         int,
        sort_ngb_cells:                 bool,
        exclude_cell_by_dist:           bool,
    ) -> None:

        # split halos into big (pre-chcked) and small
        cells, n_halos =  celled_halos.cells, celled_halos.n_halos
        src_celled_halos = celled_halos
        assert n_halos >= 1 
        assert pre_chk_expand_policy in (0,1)
        
        pre_chk_n_big_halos = min(pre_chk_n_big_halos, n_halos-1)
        if pre_chk_n_big_halos > 0:
            x, r = celled_halos.x, celled_halos.r
            args = np.argsort(r)
            
            n_small_halos = n_halos - pre_chk_n_big_halos
            args_small = args[:n_small_halos]
            args_big = args[n_small_halos:]
            
            celled_halos = _CelledHalos(cells, x[args_small], r[args_small])
            big_halos = _CelledHalos(
                _Cells(1, cells.l_box), x[args_big], r[args_big])
            hid_map = np.concatenate((
                args_small[celled_halos.sort_args],
                args_big[big_halos.sort_args]
            ))
        else:
            big_halos = _CelledHalos(_Cells(1, cells.l_box), None, None)
            hid_map = np.arange(n_halos, dtype=np.int64)
            
        self.src_celled_halos = src_celled_halos
        self.celled_halos = celled_halos 
        self.big_halos = big_halos
        self.hid_map = hid_map
        
        self.brute_force = brute_force
        
        self.pre_chk_r_eff_max = pre_chk_r_eff_max
        self.pre_chk_n_big_halos = pre_chk_n_big_halos
        self.pre_chk_expand = pre_chk_expand
        self.pre_chk_expand_policy = pre_chk_expand_policy
        self.pre_chk_expand_r_eff_min = pre_chk_expand_r_eff_min
        self.pre_chk_expand_r_min = pre_chk_expand_r_min
        
        # create look-up cells for max halo radius
        if not use_celled_r_h_max:
            ch = None
        else:
            assert celled_r_h_max_n_grids > 0
            ch = celled_halos.create_new_n_grids(celled_r_h_max_n_grids)
        celled_r_h_max = _CelledMaxHaloRadii(ch)
        
        self.r_max = r_max
        self.celled_r_h_max = celled_r_h_max
        self.use_celled_r_h_max = use_celled_r_h_max
        self.celled_r_h_max_n_grids = celled_r_h_max_n_grids
        self.sort_ngb_cells = sort_ngb_cells
        self.exclude_cell_by_dist = exclude_cell_by_dist
        
    def locate(self, x: np.ndarray):
        '''
        Return: 
        - Index into cell_halos (x, r, etc) passed to __init__().
        - r_eff to the closest halo.
        '''
        x = x.astype(np.float64)
        if self.brute_force:
            hid, r_eff_sq = self.brute_force_chk(x)
        else:
            cid_3d, hid, r_eff_sq, expanded = self.pre_chk(x)
            hid, r_eff_sq = self.post_chk(x, cid_3d, hid, r_eff_sq, expanded)
            if hid >= 0:
                hid = self.hid_map[hid]
        r_eff = np.sqrt(r_eff_sq)
        return  hid, r_eff
        
    def brute_force_chk(self, x: np.ndarray):
        ch = self.src_celled_halos
        n_halos, cells, xs, r_sq_hs = ch.n_halos, ch.cells, ch.x, ch.r_sq
        hid, r_eff_sq = -1, self.pre_chk_r_eff_max**2
        for i in range(n_halos):
            _r_eff_sq = cells.point_dist(x, xs[i]) / r_sq_hs[i]
            if _r_eff_sq < r_eff_sq:
                hid = i
                r_eff_sq = _r_eff_sq
        return hid, r_eff_sq
        
    def pre_chk(self, x: np.ndarray):
        c_halos = self.celled_halos
        
        cells = c_halos.cells
        cid_3d = cells.x_to_cid_3d(x)
        cid_1d = cells.cid_3d_to_1d(cid_3d)

        hid, r_eff_sq = -1, self.pre_chk_r_eff_max**2
        hid, r_eff_sq = c_halos.min_dist_to(x, cid_1d, hid, r_eff_sq)
        
        # check for big halos one by one
        if self.pre_chk_n_big_halos > 0:
            _hid, r_eff_sq = self.big_halos.min_dist_to_all(x, hid, r_eff_sq)
            if _hid != hid:
                hid = _hid + c_halos.n_halos
                
        # expand search radius for small halos
        expanded = False
        if self.pre_chk_expand:
            r_sq = self.__r_sq_from_eff(hid, r_eff_sq) if hid >= 0 \
                else self.r_max**2
            if r_sq >= self.pre_chk_expand_r_min**2 \
                or r_eff_sq >= self.pre_chk_expand_r_eff_min**2:
                hid, r_eff_sq = c_halos.min_dist_to_expanded(
                    self.pre_chk_expand_policy, x, cid_3d, hid, r_eff_sq)
                expanded = True
        
        return cid_3d, hid, r_eff_sq, expanded
    
    def post_chk(self, x: np.ndarray, cid_3d: np.ndarray,
        hid: int, r_eff_sq: float, expanded: bool):

        r_sq_max, c_halos, c_r_h_max = (self.r_max**2, 
            self.celled_halos, self.celled_r_h_max)
        r_h_sq_max = c_halos.r_sq_max
        r_sq_max = min(r_eff_sq * r_h_sq_max, r_sq_max)
        r_max = np.sqrt(r_sq_max)
        
        # try to shrink r_max use celled_r_h_max
        if self.use_celled_r_h_max:
            r_h_sq_max = c_r_h_max.near(x, r_max)
            r_sq_max = min(r_eff_sq * r_h_sq_max, r_sq_max)
            r_max = np.sqrt(r_sq_max)
            
        cells = c_halos.cells
        lbs, ubs = cells.x_to_cid_3d(x - r_max), \
            cells.x_to_cid_3d(x + r_max) + 1
        
        if not self.sort_ngb_cells:
            for i0 in range(lbs[0], ubs[0]): 
                for i1 in range(lbs[1], ubs[1]):
                    for i2 in range(lbs[2], ubs[2]):
                        ngb_cid_3d = np.array((i0,i1,i2), dtype=np.int64)
                        ngb_cid_1d = cells.cid_3d_to_1d(ngb_cid_3d)
                        if not self.__cell_needs_chk(x, ngb_cid_3d, ngb_cid_1d, 
                            cid_3d, expanded, r_sq_max): continue
                        hid, r_eff_sq = c_halos.min_dist_to(x, ngb_cid_1d, 
                            hid, r_eff_sq)
                        r_sq_max = min(r_eff_sq * r_h_sq_max, r_sq_max) 
            return hid, r_eff_sq
        
        for i0 in self.__sort_cids(cid_3d[0], lbs[0], ubs[0]):
            for i1 in self.__sort_cids(cid_3d[1], lbs[1], ubs[1]):
                for i2 in self.__sort_cids(cid_3d[2], lbs[2], ubs[2]):
                    ngb_cid_3d = np.array((i0,i1,i2), dtype=np.int64)
                    ngb_cid_1d = cells.cid_3d_to_1d(ngb_cid_3d)
                    if not self.__cell_needs_chk(x, ngb_cid_3d, ngb_cid_1d, 
                        cid_3d, expanded, r_sq_max): continue
                    hid, r_eff_sq = c_halos.min_dist_to(x, ngb_cid_1d, 
                        hid, r_eff_sq)
                    r_sq_max = min(r_eff_sq * r_h_sq_max, r_sq_max) 
                    
        return hid, r_eff_sq
        
    def __r_sq_from_eff(self, hid, r_eff_sq):
        c_halos = self.celled_halos
        if hid >= c_halos.n_halos:
            r_sq = r_eff_sq * self.big_halos.r_sq[hid-c_halos.n_halos]
        else:
            r_sq = r_eff_sq * c_halos.r_sq[hid]
        return r_sq        
              
    def __cell_needs_chk(self, x: np.ndarray, 
        cid_3d: np.ndarray, cid_1d: int, x_cid_3d: np.ndarray, 
        expanded: bool, r_sq_max: float):
        
        c_halos = self.celled_halos
        if c_halos.cell_sizes[cid_1d] == 0:
            return False
        
        # try to exclude cells that had been visited in pre_chk().
        d_cid = cid_3d - x_cid_3d
        if self.pre_chk_expand and expanded:
            if self.pre_chk_expand_policy == 0:
                if np.abs(d_cid).sum() <= 1: 
                    return False
            elif (np.abs(d_cid) <= 1).all():
                    return False
        elif not d_cid.any():
            return False
        
        # try to exclude distant cells
        if self.exclude_cell_by_dist:
            r_sq = c_halos.cells.min_dist_to_cell(x, x_cid_3d, cid_3d)
            if r_sq >= r_sq_max:
                return False
            
        return True
    
    def __sort_cids(self, cid: int, lb: int, ub: int):
        out_cids = np.arange(lb, ub)
        d_cids = np.abs(out_cids - cid)
        args = np.argsort(d_cids)
        return out_cids[args]


class DomainLocatorDetail(abc_e.NumbaClassWrapper):
    Impl = _DomainLocatorDetail
    impl_repr_items = (
        ('src_celled_halos', CelledHalos),
        ('celled_halos', CelledHalos), 
        ('big_halos', CelledHalos),
        'brute_force',
        'pre_chk_r_eff_max', 
        'pre_chk_n_big_halos', 
        'pre_chk_expand', 
        'pre_chk_expand_policy',
        'pre_chk_expand_r_eff_min', 
        'pre_chk_expand_r_min',
        'r_max', 
        ('celled_r_h_max', CelledMaxHaloRadii),
        'use_celled_r_h_max', 
        'celled_r_h_max_n_grids',
        'sort_ngb_cells', 
        'exclude_cell_by_dist',
    )


@dataclass
class Policy:
    n_grids:                        int     = 16
    brute_force:                    bool    = False
    pre_chk_r_eff_max:              float   = 100.0
    pre_chk_n_big_halos:            int     = 16
    pre_chk_expand:                 bool    = True    
    pre_chk_expand_policy:          int     = 0
    pre_chk_expand_r_eff_min:       float   = 10.0
    pre_chk_expand_r_min:           float   = 10.0
    r_max:                          float   = 100.0
    use_celled_r_h_max:             bool    = True
    celled_r_h_max_n_grids:         int     = 8
    sort_ngb_cells:                 bool    = False
    exclude_cell_by_dist:           bool    = True

    def create_detail(self, celled_halos: CelledHalos):
        
        ch: _CelledHalos = celled_halos.impl
        assert ch.cells.n_grids == self.n_grids
        detail = DomainLocatorDetail.create(
            celled_halos            = ch,
            **self.__detail_kw
        )
        return detail
    
    @property
    def __detail_kw(self) -> dict:
        keys = (
            'brute_force', 
            'pre_chk_r_eff_max', 'pre_chk_n_big_halos', 'pre_chk_expand', 
            'pre_chk_expand_policy', 'pre_chk_expand_r_eff_min', 
            'pre_chk_expand_r_min', 'r_max', 'use_celled_r_h_max',
            'celled_r_h_max_n_grids', 'sort_ngb_cells',
            'exclude_cell_by_dist',)
        return { k: getattr(self, k) for k in keys }


@dataclass
class LocatePolicy:
    pass

class DomainLocator(abc.HasLog, abc.HasSimpleRepr):
    
    Detail          = DomainLocatorDetail
    Policy          = Policy
    LocatePolicy    = LocatePolicy
    
    def __init__(self, halo_cat: SimHaloCatalog, 
                 m_key = 'm_mean200', pl = None, **kw) -> None:
        '''
        @halo_cat: 'x', m_key, and r_halo() are used. 'x' is shifted into the 
        box before constructing the locator.
        '''
        
        super().__init__(**kw)
    
        if pl is None:
            pl = Policy()
            
        celled_halos = self.__take_halos(halo_cat, m_key, pl)
        detail = pl.create_detail(celled_halos)
        
        self.m_key, self.pl  = m_key, pl
        self.celled_halos = celled_halos
        self.detail = detail
    
    def to_simple_repr(self) -> dict:
        return {
            'm_key':    self.m_key,
            'pl':       self.pl,
            'detail':   self.detail.to_simple_repr(),
        }
        
    def locate(self, x: np.ndarray, pl: LocatePolicy = None):
        '''
        @x: np.ndarray, shape=(n_points, 3).
            Each of x can be slightly outside the box for which a periodic 
            boundary condition is adopted.
        Return:
        - hids: indices into matched halos in halo_cat of __init__(). 
          -1 for no matching.
        - r_effs: effective distances to the matched halos.
        '''
        
        if pl is None:
            pl = LocatePolicy()
        
        detail, sort_args = self.detail.impl, self.celled_halos.impl.sort_args
        out = self.__locate(x, detail, sort_args)        
        return out
        
    @staticmethod
    def __take_halos(halo_cat: SimHaloCatalog, m_key: str, pl: Policy):
        assert m_key == 'm_mean200'
        
        sim_info = halo_cat.sim_info
        dt = np.float64
        x = halo_cat['x'].astype(dt)
        r = halo_cat.r_halo('mean200').astype(dt)
        
        cells = Cells.create(n_grids=pl.n_grids, l_box=sim_info.box_size)
        DomainLocator.__shift_xs_in(x, cells.impl)
        out = CelledHalos.create(cells=cells, x=x, r=r)
        return out
            
    @staticmethod
    @njit
    def __shift_xs_in(xs: np.ndarray, cells: _Cells):
        n_points = len(xs)
        for i in range(n_points):
            cells.shift_x_in(xs[i])
    
    @staticmethod
    @njit
    def __locate(xs: np.ndarray, detail: _DomainLocatorDetail, 
                 sort_args: np.ndarray):
        n_points = len(xs)
        hids = np.empty(n_points, dtype=np.int64)
        r_effs = np.empty(n_points, dtype=np.float64)
        for i, _x in enumerate(xs):
            hid, r_eff = detail.locate(_x)
            if hid >= 0:
                hid = sort_args[hid]
            hids[i] = hid
            r_effs[i] = r_eff
        return hids, r_effs
    
    @staticmethod
    @njit(parallel=True)
    def __locate_parallel(xs: np.ndarray, detail: _DomainLocatorDetail):
        n_points = len(xs)
        hids = np.empty(n_points, dtype=np.int64)
        r_effs = np.empty(n_points, dtype=np.float64)
        for i in prange(n_points):
            _x = xs[i]
            hid, r_eff = detail.locate(_x)
            hids[i] = hid
            r_effs[i] = r_eff
        return hids, r_effs


from __future__ import annotations
import typing
from typing import Self
from pyhipp_sims import sims
from pyhipp.core import abc
import numpy as np


class SimFrame(abc.HasDictRepr):

    repr_attr_keys = 'sim_info',

    def __init__(self, sim_info: sims.SimInfo | str, **base_kw):
        super().__init__(**base_kw)
        if isinstance(sim_info, str):
            sim_info = sims.predefined[sim_info]
        self.sim_info: sims.SimInfo = sim_info


class ComaLikeObsFrame(SimFrame):

    repr_attr_keys = SimFrame.repr_attr_keys + ('coma_pos_j2k',)

    def __init__(self, coma_pos_j2k: np.ndarray, **base_kw) -> None:
        super().__init__(**base_kw)

        coma_pos_j2k = np.array(coma_pos_j2k, dtype=float)
        coma_axes = self.projected_axes(coma_pos_j2k)

        self.coma_pos_j2k = coma_pos_j2k
        self.coma_axes = coma_axes

    @staticmethod
    def projected_axes(pos_target):
        norm = np.linalg.norm(pos_target)
        ax_z = pos_target / norm

        tmp_ax_y = np.array([0.0, 0.0, 1.0], dtype=float)

        ax_x = np.cross(tmp_ax_y, ax_z)
        ax_x /= np.linalg.norm(ax_x)

        ax_y = np.cross(ax_z, ax_x)

        return np.array([ax_x, ax_y, ax_z])


class SimFrameSdssL500(ComaLikeObsFrame):

    repr_attr_keys = ComaLikeObsFrame.repr_attr_keys + (
        'earth_pos_sim', 'phi0', 'j2k_axes', 'earth_axes',)

    def __init__(self, sim_info: sims.SimInfo | str = 'elucid',
                 coma_pos_j2k: np.ndarray = [-61.61415, -16.29136, 33.81053],
                 **base_kw):

        super().__init__(coma_pos_j2k=coma_pos_j2k,
                         sim_info=sim_info, **base_kw)

        phi0 = 39.0 * np.pi/180.0
        j2k_axes = np.array([
            [np.cos(phi0), np.sin(phi0), 0.0],
            [np.cos(phi0 + np.pi/2.0), np.sin(phi0 + np.pi/2.0), 0.0],
            [0.0, 0.0, 1.0]
        ], dtype=float)
        earth_axes = np.array([
            [np.cos(-phi0), np.sin(-phi0), 0.0],
            [np.cos(np.pi/2.0 - phi0), np.sin(np.pi/2.0 - phi0), 0.0],
            [0.0, 0.0, 1.0]], dtype=float)

        self.earth_pos_sim = np.array([370.0, 370.0, 30.0], dtype=float)
        self.phi0 = phi0
        self.j2k_axes = j2k_axes
        self.earth_axes = earth_axes

    def rebind_coma(self, coma_pos_j2k: np.ndarray) -> Self:
        return self.__class__(sim_info=self.sim_info,
                              coma_pos_j2k=coma_pos_j2k)

    def vel_sim_to_j2k(self, vel_sim: np.ndarray) -> np.ndarray:
        return np.matmul(vel_sim, self.j2k_axes.T)
    
    def los_vel(self, pos_sim: np.ndarray, vel_sim: np.ndarray) -> np.ndarray:
        pos, vel = self.pos_sim_to_j2k(pos_sim), self.vel_sim_to_j2k(vel_sim)
        return self.los_vel_from_j2k(pos, vel)
    
    def los_vel_from_j2k(self, x_j2k, v_j2k):
        x_unit = x_j2k / np.linalg.norm(x_j2k, axis=1, keepdims=True)
        v_los = (v_j2k * x_unit).sum(axis=1)
        return v_los

    def pos_sim_to_j2k(self, pos_sim: np.ndarray) -> np.ndarray:
        pos = pos_sim - self.earth_pos_sim
        return np.matmul(pos, self.j2k_axes.T)

    def pos_j2k_to_sim(self, pos_j2k: np.ndarray) -> np.ndarray:
        pos = np.matmul(pos_j2k, self.earth_axes.T)
        return pos + self.earth_pos_sim

    def pos_j2k_to_ra_dec_z(
            self, pos_j2k,
            z_finder='interp',
            interp_kw={'dz': 1.0e-2}) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        '''
        Return ra, dec and z.
        @z_finder: 'interp' or 'brute-force'
        '''
        assert z_finder in ['interp', 'brute-force']

        norm = np.linalg.norm(pos_j2k, axis=1, keepdims=True)
        norm[norm < 1.0e-10] = 1.0e-10

        pos_j2k_normed = pos_j2k / norm
        ra, dec = self.pos_spherical_to_ra_dec(pos_j2k_normed)

        if z_finder == 'interp':
            z = self.interp_cdis_to_z(self.sim_info, norm[:, 0], **interp_kw)
        else:
            z = self.sim_info.cosmology.redshifts.at_comoving(norm[:, 0])

        return ra, dec, z
    
    def pos_ra_dec_d_to_j2k(self, ra: np.ndarray, dec: np.ndarray,
                            d_c: np.ndarray) -> np.ndarray:
        j2k_u = self.pos_ra_dec_to_spherical(ra, dec)
        j2k = j2k_u * d_c[:, None]
        return j2k

    def pos_ra_dec_z_to_j2k(self, ra: np.ndarray, dec: np.ndarray,
                            z: np.ndarray) -> np.ndarray:
        j2k_u = self.pos_ra_dec_to_spherical(ra, dec)
        d = self.sim_info.cosmology.distances.comoving_at(z)
        j2k = j2k_u * d[:, None]
        return j2k

    def pos_j2k_to_coma(self, pos_j2k: np.ndarray) -> np.ndarray:
        pos = pos_j2k - self.coma_pos_j2k
        return np.matmul(pos, self.coma_axes.T)

    @staticmethod
    def interp_cdis_to_z(sim_info: sims.SimInfo, d: np.ndarray,
                         dz: float = 1.0e-2):
        assert d.min() >= 0

        cosm = sim_info.cosmology
        z_max = cosm.redshifts.at_comoving(d.max()) + 0.1
        zp = np.arange(0.0, z_max, step=dz, dtype=float)
        dp = cosm.distances.comoving_at(zp)

        z = np.interp(d, dp, zp)
        return z

    @staticmethod
    def pos_spherical_to_ra_dec(pos_spherical):
        '''
        Find RA, Dec in degree.
        norm of ``pos_spherical`` must be 1.0 in each row.
        '''

        dec = np.arcsin(pos_spherical[:, 2])
        ra = np.arctan2(pos_spherical[:, 1], pos_spherical[:, 0])

        sel = ra < 0.
        ra[sel] = ra[sel] + 2.0*np.pi

        ra *= 180.0 / np.pi
        dec *= 180.0 / np.pi

        return ra, dec

    @staticmethod
    def pos_ra_dec_to_spherical(ra: np.ndarray, dec: np.ndarray):

        ra, dec = ra * np.pi / 180.0, dec * np.pi / 180.0

        r = np.cos(dec)
        x0 = r * np.cos(ra)
        x1 = r * np.sin(ra)
        x2 = np.sin(dec)
        return np.column_stack([x0, x1, x2])



from pyhipp.astro.cosmology.model import LambdaCDM
from dataclasses import dataclass
import numpy as np


class LinearPerturbationTheory:
    
    @dataclass
    class Policy:
        growth_rate_f_approx: str     = 'Lahav1991'   # 'Peebles1980', 'Lahav1991'

    
    def __init__(self, cosm: LambdaCDM, policy: Policy = None) -> None:
        self.cosm = cosm
        
        if policy is None:
            policy = LinearPerturbationTheory.Policy()
        self.policy = policy
    
    def growth_rate_f(self, z: np.ndarray) -> np.ndarray:
        cosm, pl = self.cosm, self.policy
        o_m = cosm.omega_m(z)
        f = o_m ** 0.6
        if pl.growth_rate_f_approx == 'Lahav1991':
            o_l = cosm.omega_l(z)
            f += o_l / 70. * (0.5 * o_m + 1.)
        elif pl.growth_rate_f_approx == 'Peebles1980':
            pass
        else:
            raise KeyError(f'Unknown growth rate approximation: '
                           f'{pl.growth_rate_f_approx}')
        return f
    


from __future__ import annotations
import numpy as np
from numba import njit, int64, float64, prange
from numba.experimental import jitclass
from pyhipp.core import abc, dataproc as dp
from .periodic import PeriodicField
from ..fft import FFTPolicy, default_fft_policy

class Smoother(abc.HasDictRepr, abc.HasLog):
    
    Field = PeriodicField
    
    def run(self, field: Field) -> Field:
        '''
        @field: has to be density contrast field.
        '''
        raise NotImplementedError()
    
    @property
    def volume(self) -> float:
        raise NotImplementedError()

class TopHatSmoother(Smoother):
    
    Field = Smoother.Field    
    repr_attr_keys = 'r', 'filter_shape'
    
    def __init__(self, r: float, filter_shape='cubic', **kw) -> None:
        super().__init__(**kw)
        
        assert filter_shape in ('cubic', )
        
        self.r = r
        self.filter_shape = filter_shape
        
    @property
    def volume(self) -> float:
        return (2.0*self.r) ** 3
        
    def run(self, field: Field):
        return self.__run_cubic(field)
        
    def __run_cubic(self, field: Field):
        ccfg, data = field.cell_cfg, field.data
        l_grid, n_grids = ccfg.l_grid, ccfg.n_grids
        
        n_sm = int(np.ceil(self.r / l_grid))
        assert n_sm < n_grids
        self.log(f'Running with cubic filter, l={self.r}, '
                 f'using {n_sm} neighbors')

        data_out = np.empty_like(data)
        self.__run_cubic_impl(data, n_sm, data_out)
        return TopHatSmoother.Field(data=data_out, cell_cfg=ccfg)
     
    @staticmethod
    def __run_cubic_impl(data: np.ndarray, n_sm: int, data_out: np.ndarray):
        n = data.shape[0]
        scale = 1.0 / (2 * n_sm + 1) ** 3
        for i0 in range(n):
            for i1 in range(n):
                for i2 in range(n):
                    x_all = 0.
                    for d0 in range(-n_sm, n_sm+1):
                        for d1 in range(-n_sm, n_sm+1):
                            for d2 in range(-n_sm, n_sm+1):
                                _i0 = (i0 + d0) % n
                                _i1 = (i1 + d1) % n
                                _i2 = (i2 + d2) % n
                                x_all += data[_i0,_i1,_i2]
                    data_out[i0,i1,i2] = x_all * scale
                    
class FFTTopHatSmoother(Smoother):
    
    Field = Smoother.Field

    repr_attr_keys = 'r'

    def __init__(self, r: float, fft_pl: FFTPolicy = default_fft_policy, 
                 **kw) -> None:
        
        super().__init__(**kw)
        
        self.r = r
        self.fft_pl = fft_pl
        
    @property
    def volume(self) -> float:
        return 4./3. * np.pi * self.r ** 3
    
    def run(self, field: Field) -> Field:
        fft_pl = self.fft_pl
        f_in_k = field.fft(fft_pl)
        return self.run_from_k_space(f_in_k)
    
    def run_from_k_space(self, f_in_k: Field) -> Field:
        fft_pl = self.fft_pl
        ccfg = f_in_k.cell_cfg
        l, l_box, n = self.r, ccfg.l_box, ccfg.n_grids
        k_scale = 2.0 * np.pi * l / l_box
        n_z = n // 2 + 1
        
        assert f_in_k.shape == (n, n, n_z)
        self.__mul_filter_in_k(f_in_k.data, k_scale)
        f_in_x = f_in_k.ifft(fft_pl)
        
        data = f_in_x.data
        norm = data.mean()
        data /= norm
        if np.abs(norm - 1.) > 1.0e-2:
            self.log(f'Normalized real space field by {norm}')
        
        return f_in_x
    
    @staticmethod
    @njit
    def __mul_filter_in_k(f_in_k: np.ndarray, k_scale: float):
        n0, n1, n2 = f_in_k.shape
        n_half = n0 // 2
        for i0 in prange(n0):
            _i0 = np.float64(i0 if i0 < n_half else i0 - n0)
            for i1 in range(n1):
                _i1 = np.float64(i1 if i1 < n_half else i1 - n1)
                for i2 in range(n2):
                    _i2 = np.float64(i2)
                    if i0 == i1 == i2 == 0:
                        continue
                    kl = np.sqrt(_i0*_i0 + _i1*_i1 + _i2*_i2) * k_scale
                    kl3 = kl*kl*kl
                    w = 3.0 * (np.sin(kl) - kl * np.cos(kl)) / kl3
                    f_in_k[i0, i1, i2] *= w

class FFTMassAdaptiveSmoother(Smoother):
    
    Field = Smoother.Field
    repr_attr_keys = 'mean_density', 'target_mass', 'lagrangian_r', \
        'dr_rel', 'dr_abs', 'r_max', 'fft_pl'
    
    def __init__(self, mean_density, target_mass, 
        dr_rel = 0.2, dr_abs = 0.1, r_max = 32.0,
        fft_pl: FFTPolicy = default_fft_policy, **kw) -> None:
        
        super().__init__(**kw)
        
        r3 = target_mass / (4.0/3.0*np.pi*mean_density)
        r = r3**(1./3.)
        
        self.mean_density = mean_density
        self.target_mass = target_mass
        self.lagrangian_r = r
        self.dr_rel = dr_rel
        self.dr_abs = dr_abs
        self.r_max = r_max
        self.fft_pl = fft_pl
        
    def run(self, field: Field) -> Field:
        ccfg = field.cell_cfg
        r, dr_rel, dr_abs, r_max = ccfg.l_grid, self.dr_rel, \
            self.dr_abs, self.r_max
        
        self.log(f'Start smoothing with {r=}')
        f_in_k = field.fft(self.fft_pl)
        out_f = self.__get_smoothed(f_in_k, r)
        c = self.__chk_convergence(out_f, r, None)
        left_idx = (c < 0.0).nonzero()[0]
        c = c[left_idx]
        
        while left_idx.size > 0:
            next_r = max(r*(1.+dr_rel), r + dr_abs)
            if next_r > r_max:
                self.log(f'r ({next_r}) > maximal r ({r_max}). Iteration ends.')
                break
            self.log(f'Try {r=} for {left_idx.size} grids, mean mass ratio '
                     f'= {c.mean()}')
            next_f = self.__get_smoothed(f_in_k, next_r)
            next_c = self.__chk_convergence(next_f, next_r, left_idx)
            left_idx, c = self.__update_field(
                out_f, next_f, left_idx, c, next_c)
            r = next_r

        data = out_f.data
        norm = data.mean()
        data /= norm
        if np.abs(norm - 1.) > 1.0e-2:
            self.log(f'Normalized real space field by {norm}')

        return out_f

    def __get_smoothed(self, f_in_k: Field, r: float):
        s = FFTTopHatSmoother(r, fft_pl=self.fft_pl, verbose=self.verbose)
        return s.run_from_k_space(f_in_k)
    
    def __chk_convergence(self, f_in_x: Field, r: float, idx: np.ndarray):
        d_target = self.target_mass / ( 
            4.0 / 3.0 * np.pi * r ** 3 * self.mean_density)
        d = f_in_x.data.ravel()
        if idx is not None:
            d = d[idx]
        c = dp.Num.safe_lg(d / d_target)
        return c

    @staticmethod
    def __update_field(f: Field, next_f: Field, idx: np.ndarray, 
            c: np.ndarray, next_c: np.ndarray):
        d, next_d = f.data.ravel(), next_f.data.ravel()
        left = next_c < 0.0
        conv = ~left
        conv_idx, left_idx = idx[conv], idx[left]
        
        # make interpolation for the converged
        y, next_y = dp.Num.safe_lg(d[conv_idx], next_d[conv_idx]) 
        x, next_x = c[conv], next_c[conv]
        dy, dx = next_y - y, next_x - x
        assert (x < 0.0).all()
        assert (dx > 0.0).all()
        interp_y = dy / dx * (-x) + y
        d[conv_idx] = 10.0**interp_y
        
        d[left_idx] = next_d[left_idx]
    
        return left_idx, next_c[left]

@jitclass
class _MassAdaptiveSmootherCubic:
    
    data: float64[:, :, :]
    rho_sum: float64
    data_out: float64[:, :, :]
    n: int64
    
    def __init__(self, data: np.ndarray, rho_sum: float, 
                 data_out: np.ndarray) -> None:
        self.data = data
        self.rho_sum = rho_sum
        self.data_out = data_out
        self.n = data.shape[0]
        
    def run(self):
        data_out, n = self.data_out, self.n
        for i0 in range(n):
            for i1 in range(n):
                for i2 in range(n):
                    i_3d = np.array((i0, i1, i2))
                    data_out[i0, i1, i2] = self.__rho_at(i_3d)
    
    def __rho_at(self, i_3d: np.ndarray):
        data, rho_sum, n = self.data, self.rho_sum, self.n
        rho_all, n_sm = 0.0, 0
        while True:
            for d0 in range(-n_sm, n_sm+1):
                for d1 in range(-n_sm, n_sm+1):
                    for d2 in range(-n_sm, n_sm+1):
                        d_3d = np.array((d0, d1, d2))
                        if (np.abs(d_3d) < n_sm).all():
                            continue
                        i0, i1, i2 = (i_3d + d_3d) % n
                        rho_all += data[i0, i1, i2]
            if rho_all >= rho_sum:
                break
            n_sm += 1
        scale = 1.0 / (2 * n_sm + 1) ** 3
        return rho_all * scale

class MassAdaptiveSmoother(Smoother):
    
    Field = Smoother.Field
    repr_attr_keys = 'mass', 'filter_shape'
    
    def __init__(self, mass: float, filter_shape='cubic', 
                 ensure_mass_conservetive = True,
                 **kw) -> None:
        super().__init__(**kw)
        
        assert filter_shape in ('cubic', )
        
        self.mass = mass
        self.filter_shape = filter_shape
        self.ensure_mass_conservetive = ensure_mass_conservetive
        
    def run(self, field: Field):
        ccfg, data = field.cell_cfg, field.data
        l_grid = ccfg.l_grid
        vol_cell = l_grid**3
        rho_sum = self.mass / vol_cell
        
        data_out = np.empty_like(data)
        _MassAdaptiveSmootherCubic(data, rho_sum, data_out).run()
        if self.ensure_mass_conservetive:
            data_out /= data_out.mean()
        return MassAdaptiveSmoother.Field(data=data_out, cell_cfg=ccfg)
        
        



from __future__ import annotations
from numba.experimental import jitclass
import numpy as np

@jitclass
class Mesh:
    
    n_grids: int
    l_box:  float
    l_grid: float
    
    def __init__(self, n_grids: int, l_box: float) -> None:
        l_grid = l_box / n_grids
        
        self.n_grids = n_grids
        self.l_box = l_box
        self.l_grid = l_grid


import numpy as np
from pyhipp.stats.random import Rng
from . import mesh
from . import field

class Points:
    
    def __init__(self, rng: Rng = None, l_box = 100.0, v_ub = 1.0e3, 
                 mass_ub = 1.0) -> None:
    
        if rng is None:
            rng = Rng()
            
        self.rng = rng
        self.l_box = l_box
        self.v_ub = v_ub
        self.mass_ub = mass_ub
        
    def get_n(self, n = 10000):
        x = self.rng.uniform(high=self.l_box, size=(n, 3))
        v = self.rng.uniform(high=self.v_ub, size=(n, 3))
        mass = self.rng.uniform(high=self.mass_ub, size=(n,))
        return x, v, mass
    
    def get_stacked_n(self, n = 10000):
        x, v, mass = self.get_n(n)
        out = np.empty((n, 4), dtype=np.float64)
        out[:, 0] = mass
        out[:, 1:] = mass[:, None] * v
        return x, out

class MeshedPoints(Points):
    def __init__(self, n_grids = 8, **kw) -> None:
        super().__init__(**kw)
        self.n_grids = n_grids
                    
    def mesh(self):
        return mesh.Mesh(self.n_grids, self.l_box)
    
    def field(self, n_comps = None):
        if n_comps is None:
            return field.ScalarField3D(self.n_grids)
        else:
            return field.VectorField3D(self.n_grids, n_comps)


from __future__ import annotations
from typing import Any
import numpy as np
from pyhipp.core import DataDict, abc
from pyhipp.io import h5
from ..sims import SimInfo, SimInfoForElucid, predefined
from .filter import Filter

class Catalog(abc.HasSimpleRepr, abc.HasLog):
    def __init__(self, **kw) -> None:
        super().__init__(**kw)

    def __len__(self):
        raise NotImplementedError()
    
    def __getitem__(self, key):
        raise NotImplementedError()

class HaloCatalog(abc.HasSimpleRepr):
    '''
    halos/
        id, x, v, m_mean200
    meta/
        n_halos,
        type, snap, z           -- Sim
    sim_info                    -- Sim
    r_halo(key)                 -- Sim
    '''
    def __init__(self, **kw) -> None:
        super().__init__(**kw)
        
        self.halos = DataDict()
        self.meta = DataDict()
    
    @classmethod
    def create(cls, **kw):
        raise NotImplementedError()
    
    def __len__(self):
        return len(self.meta['n_halos'])
    
    def __getitem__(self, key):
        return self.halos[key]
    
    def to_simple_repr(self) -> Any:
        out = self.meta.get_dict()
        out['halos.keys'] = tuple(self.halos.keys())
        return out
        
class SimHaloCatalog(HaloCatalog):
    def __init__(self, **kw) -> None:
        super().__init__(**kw)
    
        self.sim_info: SimInfo = None
    
    @classmethod
    def create(cls, src_info: SimInfo, snap: int, **load_kw):
        info_e = SimInfoForElucid(src_info)
        
        with h5.File(info_e.root_file) as f:
            g = f[f'Snapshots/{snap}/Groups/Halos']
            halos = g.datasets.load(**load_kw)
            meta = g.attrs.load()    
        if 'id' not in halos:
            halos['id'] = np.arange(meta['n_halos'])
        meta['type'] = cls.__name__
        meta['snap'] = snap
        meta['z'] = src_info.redshifts[snap]
        
        return cls.create_from_halos(halos, meta, src_info)
    
    @classmethod
    def create_from_halos(cls, halos: DataDict, meta: DataDict, 
                          sim_info: SimInfo):
        out = cls()
        out.halos |= halos
        out.meta |= meta
        out.sim_info = sim_info
        return out
    
    def r_halo(self, key='mean200'):
        assert key == 'mean200'
        o_key = 'r_mean200'
        
        halos = self.halos
        if o_key in halos:
            return halos[o_key]
        
        ht = self.sim_info.cosmology.halo_theory
        m_halo = self['m_mean200']
        z = self.meta['z']
        rho_halo = ht.rho_vir_mean(f=200.0, z=z)
        r_halo = ht.r_vir(m_halo, rho_halo).astype(m_halo.dtype)
        halos[o_key] = r_halo
        
        return r_halo
    
    def selected(self, spec: Any, filter='and'):
        assert filter == 'and'
        args = Filter.create_and(spec)(self)
        
        halos = { k: args(v) for k, v in self.halos.items()}
        
        meta = self.meta.copy()
        meta['n_halos'] = args.selected_size
        
        return self.create_from_halos(halos, meta, self.sim_info)


from .. import gen_util as gu
from dataclasses import dataclass
import numpy as np
from pyhipp.core import abc

@dataclass
class Policy:
    
    fft_policy: gu.fft.FFTPolicy    = gu.fft.FFTPolicy(variant='real')
    mass_assignment_method: str     = 'CIC'
    
    @property
    def mass_assignment_obj(self):
        cls = gu.mass_assignment.MassAssignment
        return cls.from_method(self.mass_assignment_method)
    
class BaseField(abc.HasLog):
    def __init__(self, mesh: gu.mesh.Mesh, 
                 field_x: gu.field.Field = None, 
                 field_k: gu.field.Field = None,
                 policy: Policy = None, 
                 **kw) -> None:
        
        assert field_x is not None or field_k is not None
        if policy is None:
            policy = Policy()
        
        super().__init__(**kw)
        
        self.mesh = mesh
        self.field_x = field_x
        self.field_k = field_k
        self.policy = policy
        
    def update_field_x(self):
        self.field_x = self.field_k.fft.backward
        return self
        
    def update_field_k(self):
        self.field_k = self.field_x.fft.forward
        return self
    
    def compensate_mass_assignment_shape(self):
        '''
        Work on field_k.
        '''
        self.policy.mass_assignment_obj.compensate_shape(self.field_k)
        return self

class OverdensityField(BaseField):
        
    @staticmethod
    def create_empty(mesh: gu.mesh.Mesh, dtype: np.dtype = np.float64, 
                     policy: Policy = None):
        if policy is None:
            policy = Policy()
        f = gu.field.ScalarField3D(mesh.n_grids, dtype=dtype, 
                                   fft_policy=policy.fft_policy)
        return OverdensityField(mesh, field_x=f, policy=policy)
        
    def normalize(self):
        '''
        Update self from mass field (or number field) to over density field.
        '''
        m = self.field_x.data
        m_mean = m.mean()
        m /= m_mean
        m -= 1.
        
        return self
    
    
class PeculiarVelocityField(BaseField):
        
    @staticmethod
    def create_empty(mesh: gu.mesh.Mesh, dtype: np.dtype = np.float64, 
                     policy: Policy = None):
        if policy is None:
            policy = Policy()
        f = gu.field.VectorField3D(mesh.n_grids, dtype=dtype,
                                   fft_policy=policy.fft_policy)
        return PeculiarVelocityField(mesh, field_x=f, policy=policy)
        




from numba.experimental import jitclass
import numpy as np
from ..gen_util.mesh import Mesh

class Solver:
    pass

@jitclass
class _FFTSolverAuxiliary:
    
    mesh: Mesh
    
    def __init__(self, mesh: Mesh) -> None:
        
        self.mesh = mesh
        
    def green_fn_at_ik(self, ik: np.ndarray) -> np.ndarray:
        '''
        @ik: integer indices of the grid points in k-space. The last dimension 
        is the dimensionality of the k-space. Other dimensions are batch.
        '''
        n, h = self.mesh.n_grids, self.mesh.l_grid
        n_comps = ik.shape[-1]
        
        f0 = 2.0 / (h*h)
        f1 = np.cos((2.0*np.pi/n) * ik).sum(-1) - n_comps
        f = f0 * f1
        
        sel_0k = (ik == 0).all(-1)
        f[sel_0k] = 1.

        return np.reciprocal(f)


from __future__ import annotations
from .non_periodic import NonPeriodicField, NonPeriodicCIC
from pyhipp.core import abc, dataproc as dp
import numpy as np
import numba
import typing
from ..fft import FFTPolicy, default_fft_policy
if typing.TYPE_CHECKING:
    from .smoothers import Smoother

class CellConfig(abc.HasDictRepr):
    
    repr_attr_keys = ('n_grids', 'l_box', 'l_grid')
    
    def __init__(self, n_grids: int, l_box: float, **kw) -> None:
        super().__init__(**kw)
        
        self.n_grids = n_grids
        self.l_box = l_box
        self.l_grid = l_box / n_grids
        self.dims = (n_grids, n_grids, n_grids)
        
    @property
    def tot_n_grids(self):
        return self.n_grids ** 3
        
    def shift_x_in(self, x: np.ndarray) -> None:
        l_box = self.l_box
        x[ x >= l_box ] -= l_box
        x[ x < 0.0 ] += l_box
        
    def shift_dx(self, dx: np.ndarray) -> None:
        l_box = self.l_box
        l_half = .5 * l_box
        dx[ dx >= l_half ] -= l_box
        dx[ dx < -l_half ] += l_box
        
    def shift_x_to(self, x: np.ndarray, x_dst: np.ndarray) -> None:
        dx = x - x_dst
        self.shift_dx(dx)
        x[...] = x_dst + dx
        
    def shift_cid_in(self, cid: np.ndarray) -> None:
        n_grids = self.n_grids
        np.putmask(cid, cid < 0, cid + n_grids)
        np.putmask(cid, cid >= n_grids, cid - n_grids)
    
    def x_to_cid_no_shift(self, x: np.ndarray):
        l_grid = self.l_grid
        id = np.floor(x / l_grid).astype(int)
        return id
        
    def x_to_cid(self, x: np.ndarray) -> np.ndarray:
        id = self.x_to_cid_no_shift(x)
        n_grids = self.n_grids
        np.putmask(id, id < 0, id + n_grids)
        np.putmask(id, id >= n_grids, id - n_grids)
        return id
    
    def cid_3d_to_1d(self, cid_3d: np.ndarray) -> np.ndarray:
        ndim = cid_3d.ndim
        assert ndim in (1,2)
        if cid_3d.ndim == 2:
            cid_3d = cid_3d.T
        return np.ravel_multi_index(tuple(cid_3d), self.dims)
    
    def cid_1d_to_3d(self, cid_1d: np.ndarray):
        cid = np.unravel_index(cid_1d, self.dims)
        cid = np.array(cid)
        if cid.ndim == 2:
            cid = cid.T
        return cid
    
    def test_bound_x(self, x: np.ndarray):
        return dp.frame.PeriodicBox.test_bound(x, l_box=self.l_box)
    
    def test_bound_cid(self, cid_3d: np.ndarray):
        return ((cid_3d >= 0)&(cid_3d < self.n_grids)).all()
    
    def test_bound_rect(self, x: np.ndarray, x_range: np.ndarray):
        x_lb, x_ub = np.asarray(x_range).T
        x_cent, x_span = .5*(x_lb + x_ub), .5*(x_ub - x_lb)
        dx = x - x_cent
        self.shift_dx(dx)
        return (np.abs(dx) <= x_span).all(1)
    
class PeriodicField(NonPeriodicField):
    
    repr_attr_keys = NonPeriodicField.repr_attr_keys + ('cell_cfg',)
    
    Shape = NonPeriodicField.Shape
    DType = NonPeriodicField.DType
    Axes = NonPeriodicField.Axes
    
    def __init__(self, data: np.ndarray, 
                 spatial_n_dims: int = 3, 
                 cell_cfg: CellConfig = None,
                 **kw) -> None:
        
        super().__init__(data, spatial_n_dims, **kw)

        self.bind_cell_cfg(cell_cfg)

    @classmethod
    def from_cell_cfg(cls, cell_cfg: CellConfig, dtype: DType = np.float64,
                      fill = 0, **init_kw):
        '''
        @init_kw: such as spatial_n_dims.
        '''
        shape = cell_cfg.dims
        return cls.from_shape(shape, dtype=dtype, fill=fill, 
                              cell_cfg=cell_cfg, **init_kw)
        
    def replaced_data(self, data: np.ndarray, **init_kw):
        return PeriodicField(data=data, spatial_n_dims=self.spatial_n_dims, 
                             cell_cfg=self.cell_cfg, **init_kw)

    def copied_data(self, **init_kw):
        return self.replaced_data(self.data.copy(), **init_kw)

    def bind_cell_cfg(self, cell_cfg: CellConfig = None):
        self.cell_cfg = cell_cfg
        
    def smoothed(self, smoother: Smoother):
        return smoother.run(self)
        
    def adapt_fft_policy(self, pl: FFTPolicy = default_fft_policy):
        if pl.axes is None and self.n_dims != self.spatial_n_dims:
            pl = pl.replaced(axes=self.spatial_axes)
        return pl
           
    def fft(self, pl: FFTPolicy = default_fft_policy):
        data_out = self.adapt_fft_policy(pl).forward(self.data)
        return self.replaced_data(data_out)
        
    def ifft(self, pl: FFTPolicy = default_fft_policy):
        data_out = self.adapt_fft_policy(pl).backward(self.data)
        return self.replaced_data(data_out)
           
class MassAssigner(abc.HasDictRepr, abc.HasLog):
    
    kind: str = None
    
    def __init__(self, **kw) -> None:
        
        super().__init__(**kw)
        
    def add(self, field: PeriodicField, x: np.ndarray, 
            weight: np.ndarray = None):
        
        raise NotImplementedError()
          
@numba.experimental.jitclass
class _PeriodicCIC:
    
    n_grids:    numba.int64
    l_box:      numba.float64
    l_grid:     numba.float64
    
    def __init__(self, n_grids: int, l_box: float) -> None:
        l_grid = l_box / n_grids
        
        self.n_grids = n_grids
        self.l_box = l_box
        self.l_grid = l_grid
    
    def window_at_dx(self, dx: float):
        w = 1.0 - np.abs(dx/self.l_grid)
        if w < 0.:
            w = 0.
        return w
    
    def window_at_ik(self, ik: np.ndarray):
        '''
        @ik: integer indices of the grid points in k-space, arbitrary shaped.
        '''
        sinc = np.sinc(ik / self.n_grids)
        w = sinc * sinc
        return w
    
    
          
class PeriodicCIC(MassAssigner):
    
    kind: str = 'cic'
    Impl = _PeriodicCIC
        
    def add(self, field: PeriodicField, x: np.ndarray, 
            weight: np.ndarray = None):
        
        ccfg, data = field.cell_cfg, field.data
        l_grid, n_grids = ccfg.l_grid, ccfg.n_grids
        
        assert x.ndim == 2
        n_pts = len(x)
        if weight is not None:
            assert weight.shape == (n_pts, )
        self.log(f'Add {n_pts} points to grids')
        
        x_rel = x / l_grid
        cid_l = np.floor(x_rel).astype(np.int64)
        w_lr = (cid_l + 1) - x_rel, x_rel - cid_l
        
        # periodic boundary condition
        cid_l[ cid_l < 0 ] += n_grids
        cid_l[ cid_l >= n_grids ] -= n_grids
        assert ((cid_l >= 0)&(cid_l < n_grids)).all()
        cid_r = cid_l + 1
        cid_r[cid_r == n_grids] = 0
        cid_lr = cid_l, cid_r
        
        for i0 in 0,1:
            for i1 in 0,1:
                for i2 in 0,1:
                    cid = cid_lr[i0][:,0], cid_lr[i1][:,1], cid_lr[i2][:,2]
                    w = w_lr[i0][:, 0] * w_lr[i1][:, 1] * w_lr[i2][:, 2]
                    if weight is not None:
                        w *= weight
                    NonPeriodicCIC._add_to(data, *cid, w)
                    

    def compensate_shape(self, field: PeriodicField):
        ccfg = field.cell_cfg
        n_grids, l_box = ccfg.n_grids, ccfg.l_box
        impl = _PeriodicCIC(n_grids, l_box)
        self.__div_shape_fn(field.data, impl)
        
    @staticmethod
    @numba.njit
    def __div_shape_fn(data: np.ndarray, impl: Impl):
        n = impl.n_grids
        n0, n1, n2 = data.shape
        nd2 = n//2
        nd2p1 = nd2 + 1
        assert n2 == nd2p1 or n2 == n
        for i0 in range(n0):
            ik0 = i0 - n if i0 >= nd2 else i0
            w0 = impl.window_at_ik(ik0)
            for i1 in range(n1): 
                ik1 = i1 - n if i1 >= nd2 else i1
                w1 = impl.window_at_ik(ik1)
                for i2 in range(n2):
                    ik2 = i2 - n if i2 >= nd2 else i2
                    w2 = impl.window_at_ik(ik2)
                    w =  w0*w1*w2                        
                    data[i0,i1,i2] *= 1.0 / w


from __future__ import annotations
import typing
from typing import Self
from pyhipp.io import h5
from functools import cached_property
import numpy as np
from .peano import Peano


class SnapshotStorage:
    def __init__(self, file_name: str):
        self.file_name = file_name
        self.f = h5.File(self.file_name, 'r')
        self.dm = self.f['PartType1']

    @cached_property
    def cell_filling_exp(self):
        '''
        The exponent p, where 2**p gives the number of cells along one side of 
        the box, N_side.
        '''
        return int(self.dm['CellFillingExp'][()])

    @cached_property
    def cell_filling_curve(self):
        p = self.cell_filling_exp
        curve = Peano(p)
        return curve

    @cached_property
    def l_box(self):
        '''
        Box size, L_box.
        '''
        return self.f['BoxSize'][()]

    @cached_property
    def cell_n_ps(self):
        '''
        No of particles in each cell.
        '''
        return self.dm['NumPartInCells'][()]

    @cached_property
    def n_cells(self):
        '''
        No of cells, N_cells = N_side**3.
        '''
        return len(self.cell_n_ps)

    @cached_property
    def n_cells_side(self):
        '''
        N_side = 2**p.
        '''
        return 2**self.cell_filling_exp

    @cached_property
    def l_cell(self):
        '''
        Cell side length.
        '''
        return self.l_box / self.n_cells_side

    @cached_property
    def tot_n_ps(self):
        '''
        Total number of particles.
        '''
        return self.cell_n_ps.sum()

    @cached_property
    def cell_p_offs(self):
        '''
        Offset of the first particle in each cell.
        Sized N_cells + 1, with the last element being the total number of
        particles.
        '''
        off = np.concatenate(([0], np.cumsum(self.cell_n_ps)))
        return off

    def cell_p_off(self, i_cell: int):
        '''
        Offset of the first particle, and 1 + the offset of the last particle
        in the cell indexed i_cell (PH index).
        '''
        p_offs = self.cell_p_offs
        return p_offs[i_cell], p_offs[i_cell + 1]

    def load_range(self, b: int, e: int, key='Coordinates'):
        '''
        Load a contiguous chunk of particles globally indexed in [b, e).
        '''
        return self.dm[key][b:e]

    def cell(self, i_cell: int):
        '''
        Return a loader for the cell indexed i_cell (PH index).
        '''
        return CellStorage(self, i_cell)

    def cell_batch(self, i_cells: np.ndarray, d_max_batch=32):
        '''
        Return a loader for a batch of cells indexed ``i_cells`` (PH index).
        '''
        return CellBatchStorage(self, i_cells, d_max_batch)

    def rect_region(self, x0: np.ndarray, dx: np.ndarray):
        '''
        Return a loader for a rectangular region defined by the center x0 and
        half side length dx along each dimension. 
        
        Periodic boundary conditions are applied.
        '''
        x0, dx = np.asarray(x0), np.asarray(dx)
        
        x_lo, x_hi = x0 - dx, x0 + dx
        l_cell, n_cells_side = self.l_cell, self.n_cells_side
        i_lo, i_hi = np.floor(x_lo / l_cell).astype(int), \
            np.floor(x_hi / l_cell).astype(int)
        i0, i1, i2 = [np.arange(_l, _h+1) % n_cells_side
                      for _l, _h in zip(i_lo, i_hi)]
        i0, i1, i2 = np.meshgrid(i0, i1, i2, indexing='ij')
        cid_3 = np.column_stack((i0.ravel(), i1.ravel(), i2.ravel()))
        cid_1 = np.array(self.cell_filling_curve.peano_hilbert_keys(cid_3))

        return CellBatchStorage(self, cid_1)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.f.close()


class CellBatchStorage:
    '''
    Loader of data for a batch of cells indexed ``i_cells`` (PH index).
    
    Note that additional cells may be loaded to avoid multiple small reads.
    @d_max_batch: maximum difference in the PH index for two cells to be loaded,
    along with the cells between them, by a single call
    '''

    def __init__(self, snap_sto: SnapshotStorage, i_cells: np.ndarray,
                 d_max_batch=32):
        i_cells = np.sort(i_cells)

        self.snap_sto = snap_sto
        self.i_cells = i_cells
        self.d_max_batch = d_max_batch

    def iter(self, key='Coordinates'):
        i_cells, d_max = self.i_cells, self.d_max_batch
        c0 = c1 = i_cells[0]
        for i_cell in i_cells[1:]:
            if i_cell < c1 + d_max:
                c1 = i_cell
                continue
            yield self._load_between(c0, c1, key)
            c0 = c1 = i_cell
        yield self._load_between(c0, c1, key)

    def _load_between(self, c0, c1, key='Coordinates'):
        sto = self.snap_sto
        b = sto.cell_p_off(c0)[0]
        e = sto.cell_p_off(c1)[1]
        return sto.load_range(b, e, key)


class CellStorage:
    '''
    Loader of data for a single cell.
    '''

    def __init__(self, snap_sto: SnapshotStorage, i_cell: int):
        self.snap_sto = snap_sto
        self.i_cell = i_cell
        self.p_beg, self.p_end = snap_sto.cell_p_off(i_cell)

    @cached_property
    def x(self):
        b, e = self.p_beg, self.p_end
        return self.snap_sto.load_range(b, e, 'Coordinates')



from __future__ import annotations
from numba.experimental import jitclass
from numba import float64, int64
import numpy as np
from ..gen_util import abc as abc_e

@jitclass
class _Cells:
    
    n_grids:    int64
    l_box:      float64
    l_half:     float64
    l_grid:     float64
    n_cells:    int64
    
    def __init__(self, n_grids: int, l_box: float) -> None:
        self.n_grids = n_grids
        self.l_box = l_box
        self.l_half = l_box * .5
        self.l_grid = l_box / n_grids
        self.n_cells = n_grids * n_grids * n_grids
    
    @staticmethod
    def create_single_cell(l_box: float) -> _Cells:
        return _Cells(1, l_box)
    
    def create_new_n_grids(self, n_grids: int) -> _Cells:
        return _Cells(n_grids, self.l_box)
    
    
    def x_to_cid_3d(self, x: np.ndarray):
        return np.floor(x / self.l_grid).astype(np.int64)
    
    def x_to_cid_1d(self, x: np.ndarray):
        return self.cid_3d_to_1d(self.x_to_cid_3d(x))
    
    def cid_3d_to_1d(self, cid_3d: np.ndarray):
        n_grids = self.n_grids
        i0, i1, i2 = (cid_3d + n_grids) % n_grids
        return (i0 * n_grids + i1) * n_grids + i2
    
    def cid_3d_to_1d_unsafe(self, cid_3d: np.ndarray):
        n_grids = self.n_grids
        i0, i1, i2 = cid_3d
        return (i0 * n_grids + i1) * n_grids + i2
    
    
    def shift_dx(self, dx: np.ndarray):
        '''
        dx is updated and returned.
        '''
        l_box, l_half = self.l_box, self.l_half
        for i in range(3):
            _dx = dx[i]
            if _dx >= l_half:
                _dx -= l_box
            elif _dx < -l_half:
                _dx += l_box
            dx[i] = _dx
        return dx
    
    def shift_x_in(self, x: np.ndarray):
        '''
        x is updated and returned.
        '''
        l_box = self.l_box
        for i in range(3):
            _x = x[i]
            if _x >= l_box:
                _x -= l_box
            elif _x < 0.:
                _x += l_box
            x[i] = _x
        return x
    
    def shift_x_to(self, x: np.ndarray, x0: np.ndarray):
        '''
        Return new (shifted) x.
        '''
        return x0 + self.shift_dx(x - x0)
    
    
    def is_bound_x(self, x: np.ndarray):
        return ((0.0 <= x) & (x < self.l_box)).all()
    
    def is_bound_cid_3d(self, cid_3d: np.ndarray):
        return ((0 <= cid_3d) & (cid_3d < self.n_grids)).all()
    
    def point_dist(self, x0: np.ndarray, x1: np.ndarray) -> float:
        '''
        Return squared distance in the periodic box.
        @x0, x1: (ndim, ) ndarrays.
        '''
        dx = self.shift_dx(x1 - x0) 
        return (dx*dx).sum()
    
    @staticmethod
    def point_dist_unsafe(x0: np.ndarray, x1: np.ndarray) -> float:
        dx = x1 - x0
        return (dx*dx).sum()
    
    def min_dist_to_cell(self, x: np.ndarray, x_cid_3d: np.ndarray, 
                         cid_3d: np.ndarray):
        '''
        Return squared distance.
        '''
        r_sq, l_half, l_grid, l_box = 0., self.l_half, self.l_grid, self.l_box
        for i in range(3):
            _x, x_cid, cid = x[i], x_cid_3d[i], cid_3d[i]
            if cid < x_cid:
                _x_cell = l_grid * (cid + 1)
            elif cid > x_cid:
                _x_cell = l_grid * cid
            else:
                _x_cell = _x
            dx = _x - _x_cell
            if dx >= l_half:
                dx -= l_box
            elif dx < -l_half:
                dx += l_box
            r_sq += dx * dx
        return r_sq

class Cells(abc_e.NumbaClassWrapper):
    
    Impl = _Cells
    impl_repr_items = ('n_grids', 'l_box', 'l_grid', 'n_cells')
    
@jitclass
class _CelledHalos:
    
    cells:          _Cells
    cell_sizes:     int64[:]
    cell_offs:      int64[:]
    
    n_halos:        int64
    sort_args:      int64[:]
    x:              float64[:,:]
    r:              float64[:]
    r_sq:           float64[:]
    cid_1d:         int64[:]
    r_max:          float64
    r_sq_max:       float64
    
    
    def __init__(self, cells: _Cells, x: np.ndarray = None, 
                 r: np.ndarray = None) -> None:
        
        self.cells = cells
        if x is not None:
            assert r is not None
            self.set_halos(x, r)
            
    def create_new_n_grids(self, n_grids: int) -> _CelledHalos:
        return _CelledHalos(self.cells.create_new_n_grids(n_grids), 
                            self.x, self.r)
        
    def set_halos(self, x: np.ndarray, r: np.ndarray) -> _CelledHalos:
        
        cells = self.cells
        n_halos = len(x)
        n_cells = cells.n_cells
        assert len(r) == n_halos
        
        # find cid
        cid_1d = np.empty(n_halos, dtype=np.int64)
        for i, _x in enumerate(x):
            _cid_3d = cells.x_to_cid_3d(_x)
            assert cells.is_bound_cid_3d(_cid_3d)
            cid_1d[i] = cells.cid_3d_to_1d_unsafe(_cid_3d)
        
        # count
        cell_sizes = np.zeros(n_cells, dtype=np.int64)
        for i in cid_1d:
            cell_sizes[i] += 1
        cell_offs = np.zeros(n_cells+1, dtype=np.int64)
        for i in range(n_cells):
            cell_offs[i+1] = cell_offs[i] + cell_sizes[i]
        
        # sort
        sort_args = np.argsort(cid_1d)
        x, r, cid_1d = x[sort_args], r[sort_args], cid_1d[sort_args]
        r_max = r.max()
        r_sq = r * r
        r_sq_max = r_max * r_max
        
        self.cell_sizes = cell_sizes
        self.cell_offs = cell_offs
        
        self.n_halos = n_halos
        self.sort_args = sort_args
        self.x = x
        self.r = r
        self.r_sq = r_sq
        self.cid_1d = cid_1d
        self.r_max = r_max
        self.r_sq_max = r_sq_max
                
        return self
    
    def min_dist_to(self, x: np.ndarray, cid_1d: int, 
                    hid: int, r_eff_sq: float):
        if self.cell_sizes[cid_1d] == 0:
            return hid, r_eff_sq
        
        offs = self.cell_offs
        b, e = offs[cid_1d], offs[cid_1d+1]
        
        x_hs, r_sq_hs = self.x, self.r_sq
        cells = self.cells
        x = cells.shift_x_to(x, x_hs[b])

        for i in range(b, e): 
            _r_eff_sq = cells.point_dist_unsafe(x, x_hs[i]) / r_sq_hs[i]
            if _r_eff_sq < r_eff_sq:
                hid, r_eff_sq = i, _r_eff_sq

        return hid, r_eff_sq
    
    def min_dist_to_expanded(self, policy: int, 
        x: np.ndarray, cid_3d: np.ndarray, hid: int, r_eff_sq: float):

        if policy == 0:
            return self.__expanded_only_adjacent(x, cid_3d, hid, r_eff_sq)

        return self.__expanded_with_diagonal(x, cid_3d, hid, r_eff_sq)
    
    def __expanded_only_adjacent(self, x: np.ndarray, cid_3d: np.ndarray, 
            hid: int, r_eff_sq: float):
        for dim in range(3):
            _cid_3d = cid_3d.copy()
            for off in (-1,1):
                _cid_3d[dim] = cid_3d[dim] + off
                cid_1d = self.cells.cid_3d_to_1d(_cid_3d)
                hid, r_eff_sq = self.min_dist_to(x, cid_1d, hid, r_eff_sq)
        return hid, r_eff_sq
    
    def __expanded_with_diagonal(self, x: np.ndarray, cid_3d: np.ndarray,
            hid: int, r_eff_sq: float):
        _cid_3d = cid_3d.copy()
        for off_0 in (-1,0,1):
            _cid_3d[0] = cid_3d[0] + off_0    
            for off_1 in (-1,0,1):
                _cid_3d[1] = cid_3d[1] + off_1
                for off_2 in (-1,0,1):
                    if off_0 == off_1 == off_2 == 0:
                        continue
                    _cid_3d[2] = cid_3d[2] + off_2
                    cid_1d = self.cells.cid_3d_to_1d(_cid_3d)
                    hid, r_eff_sq = self.min_dist_to(x, cid_1d, hid, r_eff_sq)
        return hid, r_eff_sq
    
    def min_dist_to_all(self, x: np.ndarray, 
                        hid: int, r_eff_sq: float):
        cells, x_hs, r_sq_hs = self.cells, self.x, self.r_sq
        for i in range(self.n_halos):
            _r_eff_sq = cells.point_dist(x, x_hs[i]) / r_sq_hs[i]
            if _r_eff_sq < r_eff_sq:
                hid, r_eff_sq = i, _r_eff_sq
        return hid, r_eff_sq
    
    
class CelledHalos(abc_e.NumbaClassWrapper):
    Impl = _CelledHalos
    impl_repr_items = ('n_halos', 'cell_sizes', 'r', ('cells', Cells))
    quantile_repr_items = ('cell_sizes', 'r')
    
    def __init__(self, impl: Impl, **kw) -> None:
        super().__init__(impl, **kw)
        self.impl: _CelledHalos
    
@jitclass
class _CelledMaxHaloRadii:
    
    cells:          _Cells
    max_r_sq:       float64[:]
    
    def __init__(self, celled_halos: _CelledHalos = None) -> None:
        
        self.cells = _Cells(1,  1.0)
        if celled_halos is not None:
            self.set_halos(celled_halos)
        
    def set_halos(self, celled_halos: _CelledHalos) -> _CelledMaxHaloRadii:
        
        cells = celled_halos.cells
        n_cells = cells.n_cells
        max_r_sq = np.zeros(n_cells, dtype=np.float64)
        
        r_sq = celled_halos.r_sq
        cell_offs = celled_halos.cell_offs
        for i in range(n_cells):
            b, e = cell_offs[i], cell_offs[i+1]
            if e - b > 0: max_r_sq[i] = r_sq[b:e].max()   
        
        self.cells = cells
        self.max_r_sq = max_r_sq
        
        return self
            
    def near(self, x: np.ndarray, r_max: float):
        cells = self.cells
        lbs, ubs = cells.x_to_cid_3d(x - r_max), \
            cells.x_to_cid_3d(x + r_max) + 1
        r_h_sq_max = 0.
        for i0 in range(lbs[0], ubs[0]):
            for i1 in range(lbs[1], ubs[1]):
                for i2 in range(lbs[2], ubs[2]):
                    cid_3d = np.array((i0,i1,i2), dtype=np.int64)
                    cid_1d = cells.cid_3d_to_1d(cid_3d)
                    _r_h_sq_max = self.max_r_sq[cid_1d]
                    if _r_h_sq_max > r_h_sq_max:
                        r_h_sq_max = _r_h_sq_max
        return r_h_sq_max
        
class CelledMaxHaloRadii(abc_e.NumbaClassWrapper):
    
    Impl = _CelledMaxHaloRadii
    impl_repr_items = ('max_r_sq', ('cells', Cells))
    quantile_repr_items = ('max_r_sq',)